__author__ = 'bkeroack'

class InvalidKeyPairType(Exception):
    pass
class InvalidPublicKey(Exception):
    pass
class InvalidPrivateKey(Exception):
    pass

class SaltServerNotAccessible(Exception):
    pass




__author__ = 'bkeroack'

import fnmatch
import logging

import elita.util

class ValidatePermissionsObject:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, permissions):
        self.permissions = permissions

    def verify_top_dict(self):
        return isinstance(self.permissions, dict)

    def verify_top_keys(self):
        for t in self.permissions:
            try:
                assert (t == "apps") or (t == "actions") or (t == "servers")
            except AssertionError:
                return False
        return True

    def verify_perm_keys(self):
        #perm keys can be basically anything
        return True

    def verify_iterable(self):
        for t in self.permissions:
            for p in self.permissions[t]:
                try:
                    _ = (e for e in p)  # check if iterable
                except TypeError:
                    return False
        return True

    def run(self):
        if self.verify_top_dict():
            if self.verify_top_keys():
                if self.verify_perm_keys():
                    if self.verify_iterable():
                        return True
        return False


class UserPermissions:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, usersvc, token, datasvc=None):
        self.usersvc = usersvc
        self.token = token
        self.userobj = False
        self.valid_token = False
        self.datasvc = datasvc
        self.username = ""
        if self.validate_token():
            self.valid_token = True
            logging.debug("valid token")
            self.username = self.usersvc.GetUserFromToken(token)
            logging.debug("username: {}".format(self.username))

    def validate_token(self):
        return self.token in self.usersvc.GetAllTokens()

    def get_allowed_apps(self, username=None):
        '''Returns list of tuples: (appname, permissions ('read;write'))'''
        if not self.valid_token:
            return {}
        if not username:
            username = self.username
        userobj = self.usersvc.GetUser(username)
        assert self.datasvc is not None
        apps = self.datasvc.appsvc.GetApplications()
        apps.append('_global') #not returned by GetApplications()
        logging.debug("get_allowed_apps: username: {}".format(username))
        raw_app_list = [(fnmatch.filter(apps, a),
                 userobj.permissions['apps'][a]) for a in userobj.permissions['apps']]
        logging.debug("get_allowed_apps: raw_app_list: {}".format(raw_app_list))
        perms_dict = dict()
        for l in raw_app_list:  # coalesce the list, really wish this could be a dict comprehension
            perm = l[1]
            apps = l[0]
            if perm in perms_dict:
                perms_dict[perm].update(apps)
            else:
                perms_dict[perm] = set(apps)
        return {k: list(perms_dict[k]) for k in perms_dict}

    def get_allowed_actions(self, username):
        '''Returns list of tuples: (appname, actionname). If present, 'execute' permission is implicit'''
        userobj = self.usersvc.GetUser(username)
        assert self.datasvc is not None
        logging.debug("get_allowed_actions: username: {}".format(username))
        allowed_actions = dict()
        for a in userobj.permissions['actions']:
            app_list = fnmatch.filter(self.datasvc.appsvc.GetApplications(), a)
            for app in app_list:
                app_actions_allowed = set()
                for action_pattern in userobj.permissions['actions'][a]:
                    app_actions_allowed.update(tuple(fnmatch.filter(self.datasvc.jobsvc.GetAllActions(app),
                                                               action_pattern)))
                allowed_actions[app] = list(app_actions_allowed)
        return allowed_actions

    def get_allowed_servers(self, username):
        '''Returns list'''
        userobj = self.usersvc.GetUser(username)
        assert self.datasvc is not None
        logging.debug("get_allowed_servers: username: {}".format(username))
        servers = self.datasvc.serversvc.GetServers()
        return ([fnmatch.filter(servers, s) for s in userobj.permissions['servers']])

    def get_action_permissions(self, app, action):
        logging.debug("get_action_permissions: {}: {}".format(app, action))
        if self.valid_token and self.username in self.usersvc.GetUsers():
            userobj = self.usersvc.GetUser(self.username)
            if userobj.username == 'admin':
                logging.debug("returning admin permissions")
                return "execute"
            if app in userobj.permissions['actions']:
                logging.debug("{} in permissions['actions']".format(app))
                if action in userobj.permissions['actions'][app]:
                    return userobj.permissions['actions'][app][action]
                if '*' in userobj.permissions['actions'][app]:
                    return userobj.permissions['actions'][app]['*']
            if "*" in userobj.permissions['actions']:
                logging.debug("* in permissions['actions']")
                if action in userobj.permissions['actions']['*']:
                    return userobj.permissions['actions']['*'][action]
                if '*' in userobj.permissions['actions']['*']:
                    return userobj.permissions['actions']['*']['*']
            logging.debug("returning deny")
            return "deny"

    def get_app_permissions(self, app):
        logging.debug("get_permissions: app: {}".format(app))
        if self.valid_token and self.username in self.usersvc.GetUsers():
            userobj = self.usersvc.GetUser(self.username)
            if userobj.username == 'admin':
                logging.debug("returning admin permissions")
                return "read;write"
            elif "*" in userobj.permissions['apps']:
                logging.debug("found wildcard perms")
                return userobj.permissions['apps']['*']
            elif app in userobj.permissions['apps']:
                logging.debug("returning perms: {}".format(userobj.permissions['apps'][app]))
                return userobj.permissions['apps'][app]
        logging.debug("invalid user or token: {}; {}".format(self.username, self.token))
        return ""

    def validate_pw(self, username, password):
        userobj = self.usersvc.GetUser(username)
        return userobj.validate_password(password)





__author__ = 'bkeroack'

import os
import shutil
import requests
import zipfile
import tarfile
import tempfile
import logging

import elita.util

class SupportedFileType:
    TarGz = 'tar.gz'
    TarBz2 = 'tar.bz2'
    Zip = 'zip'
    types = [TarBz2, TarGz, Zip]

#async callables
def store_indirect_build(datasvc, app, build, file_type, uri, verify):
    logging.debug("indirect_upload: downloading from {}".format(uri))
    datasvc.jobsvc.NewJobData({'status': 'Downloading build file from {}'.format(uri)})
    r = requests.get(uri, verify=verify)
    fd, temp_file = tempfile.mkstemp()
    with open(temp_file, 'wb') as f:
        f.write(r.content)
    logging.debug("download and file write complete")
    datasvc.jobsvc.NewJobData({'status': "download and file write complete"})
    return store_uploaded_build(datasvc, app, build, file_type, temp_file)

def store_uploaded_build(datasvc, app, build, file_type, temp_file):
    builds_dir = datasvc.settings['elita.builds.dir']
    minimum_build_size = int(datasvc.settings['elita.builds.minimum_size'])
    bs_obj = BuildStorage(builds_dir, app, build, file_type=file_type, input_file=temp_file,
                          size_cutoff=minimum_build_size)
    datasvc.jobsvc.NewJobData({'status': 'validating file size and type'})
    if not bs_obj.validate():
        return {'error': "Invalid file type or corrupted file--check log"}

    datasvc.jobsvc.NewJobData({'status': 'storing file in builds dir'})
    fname = bs_obj.store()

    logging.debug("bs_results: {}".format(fname))

    datasvc.jobsvc.NewJobData({'status': 'updating build packages'})
    build_doc = datasvc.buildsvc.GetBuildDoc(app, build)
    build_doc['master_file'] = fname
    build_doc['packages']['master'] = {'filename': fname, 'file_type': file_type}

    for k in build_doc['packages']:
        fname = build_doc['packages'][k]['filename']
        ftype = build_doc['packages'][k]['file_type']
        found = False
        for f in build_doc['files']:
            if f['path'] == fname:
                found = True
        if not found:
            build_doc['files'].append({"file_type": ftype, "path": fname})
    build_doc['stored'] = True
    logging.debug("packages: {}".format(build_doc['packages'].keys()))
    datasvc.buildsvc.UpdateBuild(app, build, build_doc)

    datasvc.jobsvc.NewJobData({'status': 'running hook BUILD_UPLOAD_SUCCESS'})
    args = {
        'hook_parameters':
            {
                'build_name': build,
                'build_storage_info':
                    {
                      'storage_dir': bs_obj.storage_dir,
                      'filename': build_doc['master_file'],
                      'file_type': file_type
                    }
            }
    }
    res = datasvc.actionsvc.hooks.run_hook(app, 'BUILD_UPLOAD_SUCCESS', args)

    return {
        "build_stored": {
            "application": app,
            "build_name": build,
            "actions_result": res
        }
    }


class BuildError(Exception):
    pass


class BuildStorage:
    __metaclass__ = elita.util.LoggingMetaClass
    
    def __init__(self, builds_toplevel_dir=None, application=None, name=None, file_type=None, input_file=None,
                 size_cutoff=1000000):
        assert isinstance(size_cutoff, int)
        assert size_cutoff > 0
        self.builds_toplevel_dir = builds_toplevel_dir
        self.name = name
        self.application = application
        self.file_type = file_type
        self.temp_file_name = input_file
        self.size_cutoff = size_cutoff


    def create_storage_dir(self):
        build_dir = "{}{}/{}".format(self.builds_toplevel_dir, self.application, self.name)
        if not os.path.exists(build_dir):
            os.makedirs(build_dir)
        self.storage_dir = build_dir

    def store(self):
        self.create_storage_dir()
        fname = "{}/{}.{}".format(self.storage_dir, self.name, self.file_type)
        shutil.copy(self.temp_file_name, fname)
        os.remove(self.temp_file_name)
        return fname

    def validate_file_size(self):
        filesize = os.path.getsize(self.temp_file_name)
        logging.debug("validate_file_size: size_cutoff: {}".format(self.size_cutoff))
        logging.debug("validate_file_size: temp size: {}".format(filesize))
        return filesize >= self.size_cutoff

    def validate(self):
        if not self.validate_file_size():
            return False
        if self.file_type == SupportedFileType.TarGz:
            return self.validate_tgz()
        elif self.file_type == SupportedFileType.TarBz2:
            return self.validate_tbz2()
        elif self.file_type == SupportedFileType.Zip:
            return self.validate_zip()

    def validate_tgz(self):
        return self.validate_tar('gz')

    def validate_tbz2(self):
        return self.validate_tar('bz2')

    def validate_tar(self, compression):
        try:
            with tarfile.open(name=self.temp_file_name, mode='r:{}'.format(compression)) as tf:
                logging.debug("tar.{}: {}, {}, {} members".format(compression, self.application, self.name,
                                                                        len(tf.getnames())))
        except tarfile.ReadError:
            logging.debug("tar.{}: invalid tar file!".format(compression))
            return False
        return True

    def validate_zip(self):
        try:
            with zipfile.ZipFile(self.temp_file_name, mode='r') as zf:
                logging.debug("zip: {}, {}, {} members".format(self.application, self.name, len(zf.namelist())))
        except zipfile.BadZipfile:
            logging.debug("zip: invalid zip file!")
            return False
        return True


class BuildFile:
    def __init__(self, package_doc):
        self.file_type = package_doc['file_type']
        self.filename = package_doc['filename']

    def decompress(self, target_path):
        if self.file_type == SupportedFileType.Zip:
            self.decompress_zip(target_path)
        elif self.file_type == SupportedFileType.TarBz2:
            self.decompress_tbz2(target_path)
        elif self.file_type == SupportedFileType.TarGz:
            self.decompress_tgz(target_path)
        else:
            raise BuildError

    def decompress_tar(self, target_path, ext):
        with tarfile.open(name=self.filename, mode='r:{}'.format(ext)) as tf:
            tf.extractall(target_path)

    def decompress_tbz2(self, target_path):
        self.decompress_tar(target_path, 'bz2')

    def decompress_tgz(self, target_path):
        self.decompress_tar(target_path, 'gz')

    def decompress_zip(self, target_path):
        with zipfile.ZipFile(self.filename, 'r') as zf:
            zf.extractall(target_path)






from pyramid.config import Configurator
from pyramid.renderers import JSON

import pymongo

import models

def GetMongoClient(settings):
    assert settings
    client = pymongo.MongoClient(settings['elita.mongo.host'], int(settings['elita.mongo.port']), fsync=True,
                                 use_greenlets=True)
    assert client
    db = client[settings['elita.mongo.db']]
    assert db
    return db, db['root_tree'].find_one(), client

def DataStore(request):
    db, root, client = GetMongoClient(request.registry.settings)
    return db

def generate_root_tree(db):
    assert db
    tree = db['root_tree'].find_one()
    assert tree
    updater = models.RootTreeUpdater(tree, db)
    return models.RootTree(db, updater, tree, db.dereference(tree['_doc']))

def RootService(request):
    '''
    Get root tree.
    '''
    return generate_root_tree(request.db)

def DataService(request):
    return models.DataService(request.registry.settings, request.db, request.root)

def root_factory(request):
    #initialize request objects
    foo = request.db, request.datasvc
    return request.root

def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    @type settings: pyramid.registry.Registry
    """

    #data validator / migrations
    db, root, client = GetMongoClient(settings)
    dv = models.DataValidator(settings, root, db)
    dv.run()
    client.close()

    config = Configurator(root_factory=root_factory, settings=settings)
    config.add_static_view('static', 'static', cache_max_age=3600)
    config.add_renderer('prettyjson', JSON(indent=4))
    config.scan()

    config.add_request_method(DataStore, 'db', reify=True)
    config.add_request_method(RootService, 'root', reify=True)
    config.add_request_method(DataService, 'datasvc', reify=True)

    return config.make_wsgi_app()



__author__ = 'bkeroack'

import collections
import json
import sys

def is_string(obj):
    if sys.version_info[0] == 2:
        return isinstance(obj, basestring)
    elif sys.version_info[0] == 3:
        return isinstance(obj, str)

def is_seq(obj):
    return isinstance(obj, collections.Sequence) and not is_string(obj)

def is_optional_seq(obj):
    return not obj or is_seq(obj)

def is_dictlike(obj):
    return isinstance(obj, collections.Mapping)

def is_optional_dict(obj):
    return not obj or is_dictlike(obj)

def is_serializable(obj):
    '''
    Check is obj can be JSON serialized. We don't try to catch any exceptions because they generally provide better
    diagnostics
    '''
    json.dumps(obj)
    return True

def is_optional_str(obj):
    return not obj or is_string(obj)





__author__ = 'bkeroack'

from clint.textui import puts, colored

from linux import ELITA_ETC, ELITA_LOG_DIR, ELITA_HOME
from linux import do_step, create_user_and_group, mk_dir, chown_dir, cp_prod_ini_posix

def InstallOSX():

    puts("OS Flavor: Darwin/OS X")

    do_step("Creating user and group 'elita'", create_user_and_group)

    do_step("Creating log directory: {}".format(ELITA_LOG_DIR), mk_dir, [ELITA_LOG_DIR])

    do_step("Setting ownership on log directory", chown_dir, [ELITA_LOG_DIR])

    do_step("Creating config directory: {}".format(ELITA_ETC), mk_dir, [ELITA_ETC])

    do_step("Setting ownership on config directory", chown_dir, [ELITA_ETC])

    do_step("Creating running directory: {}".format(ELITA_HOME), mk_dir, [ELITA_HOME])

    do_step("Setting ownership on running directory", chown_dir, [ELITA_HOME])

    do_step("Copying ini", cp_prod_ini_posix)

    puts(colored.yellow("Elita elita.ini installed to /etc/elita/. Please run celery and pserve manually."))

    puts(colored.green("Done!"))



# Tests involving salt that must be run in the Vagrant environment

import salt.client
import multiprocessing

# def run_state_call(target, state, concurrent, q):
#     '''
#     @type q: multiprocessing.Queue
#     '''
#     sc = salt.client.LocalClient()
#     results = list()
#     ret = sc.cmd_iter(target, 'state.sls', [state], kwarg={'concurrent': concurrent})
#     for res in ret:
#         results.append(res)
#     q.put_nowait(results)
#
# def concurrent_state_calls(target, concurrent):
#     q = multiprocessing.Queue()
#     states = ['elita.scorebig.Configs', 'elita.scorebig.Artifacts']
#     results = list()
#     procs = list()
#     for s in states:
#         p = multiprocessing.Process(target=run_state_call, args=(target, s, concurrent, q))
#         p.start()
#         procs.append(p)
#
#     for p in procs:
#         p.join(300)
#
#     for r in q.get_nowait():
#         results.append(r)
#
#     print(results)
#
#
# def test_simultaneous_state_calls_concurrent_false_windows():
#     '''
#     Test simulataneous salt state calls with concurrent flag False
#     '''
#     concurrent_state_calls('web01', False)
#
# def test_simultaneous_state_calls_concurrent_true_windows():
#     '''
#     Test simulataneous salt state calls with concurrent flag True
#     '''
#     concurrent_state_calls('web01', True)
#
# if __name__ == '__main__':
#     test_simultaneous_state_calls_concurrent_false_windows()
#     test_simultaneous_state_calls_concurrent_true_windows()




__author__ = 'bkeroack'

import pymongo
import bson
import multiprocessing
import copy
import time
import random

import elita.models

def setup_db():
    mc = pymongo.MongoClient(host='localhost', port=27017)
    db = mc['elita_testing']
    return mc, db

def clear_mocks(db):
    db['mock_objs'].remove()

mc, db = setup_db()

def _create_roottree():
    root = {
        '_lock': False,
        '_doc': bson.DBRef('dummy', bson.ObjectId()),
        'app': {
            'foo': {}
        }
    }
    db['root_tree'].remove()
    db['root_tree'].insert(root)
    clear_mocks(db)

test_obj = {
    'name': 'foobar',
    'attributes': {
        'a': 0,
        'b': 1
    }
}

test_obj2 = {
    'name': 'baz',
    'attributes': {
        'a': 2,
        'b': 3
    }
}


def _threadsafe_direct_roottree_update(q, doc, path, collection, pause=0):
    ts_mc = pymongo.MongoClient(host='localhost', port=27017)
    ts_db = ts_mc['elita_testing']
    id = ts_db['mock_objs'].insert(doc)
    del doc['_id']
    path_dot_notation = '.'.join(path)
    root_tree_doc = {
        '_doc': bson.DBRef(collection, id)
    }
    time.sleep(pause)
    result = ts_db['root_tree'].update({}, {'$set': {path_dot_notation: root_tree_doc}})
    q.put(result['n'] == 1 and result['updatedExisting'] and not result['err'], block=False)


def test_get_document():
    '''
    Test that retrieving a document works
    '''
    clear_mocks(db)
    insert_doc = copy.deepcopy(test_obj)
    insert_doc['name'] = 'tuna'
    db['mock_objs'].insert(insert_doc)
    del insert_doc['_id']

    ms = elita.models.MongoService(db)
    doc = ms.get('mock_objs', {'name': 'tuna'})

    assert doc
    assert 'name' in doc
    assert doc['name'] == 'tuna'
    assert 'attributes' in doc

def test_roottree_update():
    '''
    Test that a single root_tree update does what it should
    '''
    _create_roottree()

    id = db['mock_objs'].insert(test_obj)
    del test_obj['_id']

    ms = elita.models.MongoService(db)
    ms.update_roottree(('app', 'foo', 'mocks', 'foobar'), 'mock_objs', id)

    rt_list = [d for d in db['root_tree'].find()]

    assert len(rt_list) == 1
    assert '_lock' in rt_list[0]
    assert not rt_list[0]['_lock']
    assert 'mocks' in rt_list[0]['app']['foo']
    assert 'foobar' in rt_list[0]['app']['foo']['mocks']
    assert '_doc' in rt_list[0]['app']['foo']['mocks']['foobar']
    assert rt_list[0]['app']['foo']['mocks']['foobar']['_doc'].__class__.__name__ == 'DBRef'
    assert rt_list[0]['app']['foo']['mocks']['foobar']['_doc'].collection == 'mock_objs'
    assert rt_list[0]['app']['foo']['mocks']['foobar']['_doc'].id == id


def test_roottree_direct_update():
    '''
    Test that direct modification of root_tree works
    '''

    _create_roottree()
    q = multiprocessing.Queue()

    _threadsafe_direct_roottree_update(q, test_obj, ('app', 'foo', 'mocks', 'ashley'), 'mock_objs')

    result = q.get()

    assert result
    root = [d for d in db['root_tree'].find()]
    assert len(root) == 1
    root = root[0]
    assert 'ashley' in root['app']['foo']['mocks']
    assert '_doc' in root['app']['foo']['mocks']['ashley']


def test_roottree_multiple_simultaneous_direct_updates():
    '''
    Test that multiple simultaneous root_tree updates all succeed with no data loss
    '''
    _create_roottree()
    p_list = list()
    q = multiprocessing.Queue()

    names = (
        'lecherousness',
        'undefinable',
        'agatizing',
        'jamesburg',
        'falling',
        'subtentacular',
        'acronymized',
        'chandleries',
        'croupiness',
        'rebleach',
        'unstartled',
        'unherbaceous'
    )

    doc_list = [copy.deepcopy(test_obj) for n in names]
    for n, d in zip(names, doc_list):
        d['name'] = n
        path = ('app', 'foo', 'mocks', n)
        p = multiprocessing.Process(target=_threadsafe_direct_roottree_update,
                                    args=[q], kwargs={'doc': d, 'path': path, 'collection': 'mock_objs',
                                                      'pause': random.random()/4})
        p_list.append(p)

    for p in p_list:
        p.start()

    for p in p_list:
        p.join(600)

    results = [q.get() for p in p_list]

    assert all(results)
    root = [d for d in db['root_tree'].find()]
    assert len(root) == 1
    root = root[0]
    assert all([n in root['app']['foo']['mocks'] for n in names])
    assert not root['_lock']


def test_roottree_delete_reference():
    '''
    Test that deleting a reference from root_tree works
    '''

    _create_roottree()
    insert_doc = copy.deepcopy(test_obj)
    insert_doc['name'] = 'salamander'
    ms = elita.models.MongoService(db)

    q = multiprocessing.Queue()

    _threadsafe_direct_roottree_update(q, insert_doc, ('app', 'foo', 'mocks', 'salamander'), 'mock_objs')

    rt = [d for d in db['root_tree'].find()]
    assert rt
    assert len(rt) == 1
    root = rt[0]
    assert 'app' in root
    assert 'foo' in root['app']
    assert 'mocks' in root['app']['foo']
    assert 'salamander' in root['app']['foo']['mocks']

    ms.rm_roottree(('app', 'foo', 'mocks', 'salamander'))

    rt = [d for d in db['root_tree'].find()]
    assert rt
    assert len(rt) == 1
    root = rt[0]
    assert 'app' in root
    assert 'foo' in root['app']
    assert 'mocks' in root['app']['foo']
    assert 'salamander' not in root['app']['foo']['mocks']


def test_create_new_document():
    '''
    Test that creating a new document works
    '''
    insert_doc = copy.deepcopy(test_obj)
    ms = elita.models.MongoService(db)
    res = ms.create_new('mock_objs', {'name': 'swimming'}, 'Mock', insert_doc)

    assert res
    dlist = [d for d in db['mock_objs'].find({'name': 'swimming'})]
    assert len(dlist) == 1


def test_modify_existing_document():
    '''
    Test that modifying an existing document works
    '''
    insert_doc = copy.deepcopy(test_obj)
    ms = elita.models.MongoService(db)
    res = ms.create_new('mock_objs', {'name': 'soccer'}, 'Mock', insert_doc)
    assert res

    doc = ms.get('mock_objs', {'name': 'soccer'})
    assert 'name' in doc
    assert doc['name'] == 'soccer'

    ms.modify('mock_objs', {'name': 'soccer'}, ('attributes', 'b'), 99)

    doc = ms.get('mock_objs', {'name': 'soccer'})
    assert 'name' in doc
    assert doc['name'] == 'soccer'
    assert 'attributes' in doc
    assert 'b' in doc['attributes']
    assert doc['attributes']['b'] == 99

if __name__ == '__main__':
    test_get_document()
    test_roottree_update()
    test_roottree_direct_update()
    test_roottree_multiple_simultaneous_direct_updates()
    test_roottree_delete_reference()
    test_create_new_document()
    test_modify_existing_document()



from pyramid.view import view_config
import pyramid.exceptions
import pyramid.response
import logging
import pprint
import traceback
import sys
import fnmatch
import tempfile
import json
import socket
import os

from elita.actions import action
import models
import builds
import auth
import servers
import elita.deployment.gitservice
import elita_exceptions
import elita.deployment.deploy
import elita.util

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
#lh = logging.StreamHandler(sys.stdout)
#lh.setLevel(logging.DEBUG)
#logger.addHandler(lh)

AFFIRMATIVE_SYNONYMS = ("true", "True", "TRUE", "yup", "yep", "yut", "yes", "yea", "aye", "please", "si", "sim")



class GenericView:
    #__metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, context, request, app_name="_global", permissionless=False, allow_pw_auth=False, is_action=False):
        self.required_params = {"GET": [], "PUT": [], "POST": [], "DELETE": []}  # { reqverb: [ params ] }
        logger.debug("{}: {} ; {}".format(self.__class__.__name__, context.__class__.__name__, request.subpath))
        self.req = request
        self.db = request.db
        self.datasvc = request.datasvc
        self.context = context
        if 'pretty' in self.req.params:
            if self.req.params['pretty'] in AFFIRMATIVE_SYNONYMS:
                self.req.override_renderer = "prettyjson"
        self.permissionless = permissionless
        self.is_action = is_action
        self.allow_pw_auth = allow_pw_auth
        self.setup_permissions(app_name)

    def get_created_datetime_text(self):
        return self.context.created_datetime.isoformat(' ') if hasattr(self.context, 'created_datetime') else None

    def run_async(self, name, job_type, data, callable, args):
        jid = action.run_async(self.datasvc, name, job_type, data, callable, args)
        return {
            'task': name,
            'job_id': jid,
            'status': 'async/running'
        }

    def deserialize_attributes(self):
        if 'attributes' in self.req.params and self.req.params['attributes'] in AFFIRMATIVE_SYNONYMS:
            try:
                attribs = self.req.json_body
            except:
                return False, self.Error(400, "invalid attributes object (problem deserializing, bad JSON?)")
            return True, attribs
        else:
            return True, {}

    #for verifying user-supplied parameters
    def check_against_existing(self, existing, submitted):
        return set(submitted).issubset(set(existing))

    def get_unknown(self, existing, submitted):
        return list(set(submitted) - set(existing))

    def call_action(self):
        g, p = self.check_params()
        if not g:
            return self.MISSING_PARAMETER(p)
        if self.req.method == 'POST':
            if 'execute' in self.permissions:
                return self.POST()
            else:
                return self.UNAUTHORIZED()
        else:
            return self.UNIMPLEMENTED()

    def __call__(self):
        g, p = self.check_params()
        bad_verb = False
        if not g:
            return self.MISSING_PARAMETER(p)
        if self.req.method == 'GET':
            if 'read' in self.permissions:
                return self.GET()
        elif self.req.method == 'POST':
            if 'write' in self.permissions or (self.is_action and 'execute' in self.permissions):
                return self.POST()
        elif self.req.method == 'PUT' and not self.is_action:
            if 'write' in self.permissions:
                return self.PUT()
        elif self.req.method == 'DELETE' and not self.is_action:
            if 'write' in self.permissions:
                return self.DELETE()
        else:
            bad_verb = True
        if bad_verb:
            return self.UNKNOWN_VERB()
        return self.UNAUTHORIZED()

    def check_params(self):
        try:
            for p in self.required_params[self.req.method]:
                if p not in self.req.params:
                    return False, p
        except KeyError:  # invalid HTTP verb
            pass
        return True, None

    def setup_permissions(self, app_name):
        if self.permissionless:
            self.permissions = "read;write"
            return
        if self.allow_pw_auth and 'auth_token' not in self.req.params:  # view sub-class responsible for verifying pw
            self.permissions = "read;write"
            return
        if 'auth_token' in self.req.params or 'Auth-Token' in self.req.headers:
            if 'auth_token' in self.req.params and 'Auth-Token' in self.req.headers:
                #if both are specified on one request it's an error condition
                token = ''
            else:
                submitted_tokens = self.req.params.getall('auth_token') if 'auth_token' in self.req.params else \
                    [self.req.headers['Auth-Token']]
                if len(submitted_tokens) != 1:  # multiple auth tokens are an error condition
                    token = ''
                else:
                    token = submitted_tokens[0]
            if self.is_action and self.req.method == 'POST':
                self.permissions = auth.UserPermissions(self.datasvc.usersvc, token).get_action_permissions(app_name,
                                                                                      self.context.action_name)
            else:
                self.permissions = auth.UserPermissions(self.datasvc.usersvc, token).get_app_permissions(app_name)
        else:
            self.permissions = ""

    def set_params(self, params):
        for t in params:
            self.required_params[t] = params[t]

    def return_action_status(self, action):
        return {"status": "ok", "action": action}


    def status_ok(self, msg):
        return {'status': 'ok', 'message': msg}

    def Success(self):
        return {'status': 'success'}

    def Error(self, code, message):
        body = {'status': 'error', 'error': message}
        return pyramid.response.Response(status_int=code, content_type="application/json", body=json.dumps(body))

    def GET(self):
        return self.UNIMPLEMENTED()

    def POST(self):
        return self.UNIMPLEMENTED()

    def PUT(self):
        return self.UNIMPLEMENTED()

    def DELETE(self):
        return self.UNIMPLEMENTED()

    def PATCH(self):
        return self.UNIMPLEMENTED()

    def UNKNOWN_VERB(self):
        return self.Error(501, "unknown/unsupported HTTP verb")

    def UNIMPLEMENTED(self):
        return self.Error(405, "HTTP verb '{}' not implemented for this resource".format(self.req.method))

    def UNAUTHORIZED(self):
        return self.Error(403, "insufficient permissions")

    def MISSING_PARAMETER(self, p):
        return self.Error(400, "Required parameter is missing: {}".format(p))


@view_config(context=pyramid.exceptions.HTTPNotFound, renderer='json')
class NotFoundView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, permissionless=True)
        logger.debug("REQUEST: url: {}".format(request.url))
        logger.debug("REQUEST: context: {}".format(context.__class__.__name__))
        logger.debug("REQUEST: method: {}".format(request.method))
        logger.debug("REQUEST: params: {}".format(request.params))

    def notfound(self):
        return self.Error(404, "not found (404)")
    def GET(self):
        return self.notfound()
    def POST(self):
        return self.notfound()
    def PUT(self):
        return self.notfound()
    def DELETE(self):
        return self.notfound()


@view_config(context=Exception, renderer='json')
def ExceptionView(exc, request):
    exc_type, exc_obj, tb = sys.exc_info()
    logger.exception("")
    body = {
        "status": "error",
        "error": {
            "unhandled_exception": traceback.format_exception(exc_type, exc_obj, tb)
        }
    }
    return pyramid.response.Response(status_int=500, content_type="application/json", body=json.dumps(body))

class ApplicationContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request)
        param = ["app_name"]
        self.set_params({"GET": [], "PUT": param, "POST": param, "DELETE": param})

    def validate_app_name(self, name):
        return name in self.datasvc.appsvc.GetApplications()

    def PUT(self):
        app_name = self.req.params["app_name"]
        self.datasvc.appsvc.NewApplication(app_name)
        return self.return_action_status({"new_application": {"name": app_name}})

    def DELETE(self):
        app_name = self.req.params["app_name"]
        if not self.validate_app_name(app_name):
            return self.Error(400, "app name '{}' not found".format(app_name))
        self.datasvc.appsvc.DeleteApplication(app_name)
        return self.return_action_status({"delete_application": app_name})

    def GET(self):
        return {"applications": self.datasvc.appsvc.GetApplications()}


class ApplicationView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name=context.app_name)
        self.set_params({"GET": [], "PUT": [], "POST": ["app_name"], "DELETE": []})

    def GET(self):
        return {
            "application": self.context.app_name,
            "created": self.get_created_datetime_text(),
            self.context.app_name: self.datasvc.GetAppKeys(self.context.app_name),
            "census": self.datasvc.appsvc.GetApplicationCensus(self.context.app_name)
        }

    def PATCH(self):
        pass

class ActionContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request)

    def GET(self):
        return {"application": self.context.parent,
                "actions": self.datasvc.jobsvc.GetAllActions(self.context.parent)}

class ActionView(GenericView):
    def __init__(self, context, request):
        self.app_name = context.app_name
        GenericView.__init__(self, context, request, app_name=self.app_name, is_action=True)
        self.set_params({"GET": [], "PUT": [], "POST": self.context.params, "DELETE": []})

    def execute(self):
        #we need a plain dict so we can serialize to celery
        params = {k: self.req.params[k] for k in self.req.params}
        return self.status_ok(self.context.execute(params))

    def GET(self):
        return {
            "application": self.context.app_name,
            "action": {
                "name": self.context.action_name,
                "post_parameters": self.context.details()['params']
            }
        }

    def POST(self):
        return self.execute()

class BuildContainerView(GenericView):
    def __init__(self, context, request):
        self.app_name = context.parent
        GenericView.__init__(self, context, request, app_name=self.app_name)
        self.set_params({"GET": [], "PUT": ["build_name"], "POST": ["build_name"], "DELETE": ["build_name"]})


    def validate_build_name(self, build_name):
        return build_name in self.datasvc.buildsvc.GetBuilds(self.app_name)

    def GET(self):
        return {"application": self.app_name,
                "builds": self.datasvc.buildsvc.GetBuilds(self.app_name)}

    def PUT(self):
        msg = list()
        build_name = self.req.params["build_name"]
        if '/' in build_name:
            build_name = str(build_name).replace('/', '-')
            msg.append("warning: forward slash in build name replaced by hyphen")
        if build_name in self.datasvc.buildsvc.GetBuilds(self.app_name):
            msg.append("build exists")
        attribs = self.deserialize_attributes()
        if not attribs[0]:
            return attribs[1]
        else:
            attribs = attribs[1]
        self.datasvc.buildsvc.NewBuild(self.app_name, build_name, attribs)
        return self.return_action_status({
            "new_build": {
                "application": self.app_name,
                "build_name": build_name,
                "attributes": attribs,
                "messages": msg
            }
        })

    def POST(self):
        build_name = self.req.params["build_name"]
        ok, err = self.validate_build_name(build_name)
        if not ok:
            return err
        return BuildView(self.context[build_name], self.req).POST()

    def DELETE(self):
        build_name = self.req.params["build_name"]
        if not self.validate_build_name(build_name):
            return self.Error(400, "build name '{}' not found".format(build_name))
        self.datasvc.buildsvc.DeleteBuild(self.app_name, build_name)
        return self.return_action_status({"delete_build": {"application": self.app_name, "build_name": build_name}})


class BuildView(GenericView):

    def __init__(self, context, request):
        self.app_name = context.app_name
        GenericView.__init__(self, context, request, app_name=self.app_name)
        self.set_params({"GET": [], "PUT": [], "POST": ["file_type"], "DELETE": ["file_name"]})
        self.build_name = None

    def run_build_storage(self, func, arg):
        base_args = {
            'app': self.app_name,
            'build': self.build_name,
            'file_type': self.file_type
        }
        args = dict(base_args.items() + arg.items())
        msg = self.run_async('store_build', 'async', args, func, args)
        return self.return_action_status({
            "build_stored": {
                "application": self.app_name,
                "build_name": self.build_name,
                "message": msg
            }
        })

    def run_build_storage_direct(self, tempfile):
        return self.run_build_storage(builds.store_uploaded_build, {'temp_file': tempfile})

    def run_build_storage_indirect(self, uri, verify):
        return self.run_build_storage(builds.store_indirect_build, {'uri': uri, 'verify': verify})

    def direct_upload(self):
        logger.debug("BuildView: direct_upload")
        if 'build' in self.req.POST:
            fname = self.req.POST['build'].filename
            logger.debug("BuildView: PUT: filename: {}".format(fname))
            fd, temp_file = tempfile.mkstemp()
            os.close(fd)
            with open(temp_file, 'wb') as f:
                f.write(self.req.POST['build'].file.read(-1))
            return self.run_build_storage_direct(temp_file)
        else:
            return self.Error(400, "build data not found in POST body")

    def indirect_upload(self):
        verify = self.req.params.getone('verify') in AFFIRMATIVE_SYNONYMS if 'verify' in self.req.params else False
        return self.run_build_storage_indirect(self.req.params['indirect_url'], verify)

    def POST(self):
        self.build_name = self.context.build_name
        if self.req.params["file_type"] not in models.SupportedFileType.types:
            return self.Error(400, "file type not supported")
        self.file_type = self.req.params["file_type"]

        if "indirect_url" in self.req.params:
            return self.indirect_upload()
        else:
            return self.direct_upload()

    def GET(self):
        if "package" in self.req.params:
            pkg = self.req.params['package']
            if pkg not in self.context.packages:
                return self.Error(400, "package type '{}' not found".format(pkg))
            return pyramid.response.FileResponse(self.context.packages[pkg]['filename'], request=self.req,
                                                 cache_max_age=0)
        else:
            return {'application': self.context.app_name, 'build': self.context.build_name,
                'stored': self.context.stored, 'packages': self.context.packages,
                'files': self.context.files,
                'created_datetime': self.get_created_datetime_text(),
                'attributes': self.context.attributes}


class ServerContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request)
        self.set_params({"GET": [], "PUT": ['name', 'environment', 'existing'], "POST": [], "DELETE": ['name']})

    def GET(self):
        return {
            'servers': self.datasvc.serversvc.GetServers()
        }

    def PUT(self):
        name = self.req.params['name']
        environment = self.req.params['environment']
        attribs = self.deserialize_attributes()
        existing = self.req.params['existing'] in AFFIRMATIVE_SYNONYMS
        if not attribs[0]:
            return attribs[1]
        else:
            attribs = attribs[1]
        res = self.datasvc.serversvc.NewServer(name, attribs, environment)
        if not existing:
            msg = {"error": "server provisioning not implemented yet!"}
        else:
            job_data = {
                'server': name,
                'environment': environment,
                'existing': existing
            }
            msg = self.run_async("setup_new_server", "async", job_data, servers.setup_new_server, {'name': name})
        if res['NewServer']['status'] == 'ok':
            return self.status_ok({
                "new_server": {
                    "server_name": name,
                    "environment": environment,
                    "attributes": attribs,
                    "message": msg
                }
            })
        else:
            return self.Error(500, res)

    def DELETE(self):
        name = self.req.params['name']
        self.datasvc.serversvc.DeleteServer(name)
        return self.status_ok({
            "server_deleted": {
                "server_name": name
            }
        })

class ServerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request)
        self.set_params({"GET": [], "PUT": [], "POST": [], "DELETE": []})

    def GET(self):
        return {
            'server_name': self.context.name,
            'server_type': self.context.server_type,
            'status': self.context.status,
            'created_datetime': self.get_created_datetime_text(),
            'environment': self.context.environment,
            'attributes': self.context.attributes,
            'gitdeploys': self.datasvc.serversvc.GetGitDeploys(self.context.name)
        }

    def DELETE(self):
        self.datasvc.serversvc.DeleteServer(self.context.name)
        return self.status_ok({
            "server_deleted": {
                "server_name": self.context.name
            }
        })

class EnvironmentView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request)
        self.set_params({"GET": [], "PUT": [], "POST": [], "DELETE": []})

    def GET(self):
        return {
            'environments': self.datasvc.serversvc.GetEnvironments()
        }

class DeploymentContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name=context.parent)
        self.set_params({"GET": [], "PUT": [], "POST": ['build_name'], "DELETE": []})

    def GET(self):
        return {
            'application': self.context.parent,
            'deployments': self.datasvc.deploysvc.GetDeployments(self.context.parent)
        }

    def POST(self):
        app = self.context.parent
        build_name = self.req.params['build_name']
        if build_name not in self.datasvc.buildsvc.GetBuilds(app):
            return self.Error(400, "unknown build '{}'".format(build_name))
        build_obj = self.datasvc.buildsvc.GetBuild(app, build_name)
        if not build_obj.stored:
            return self.Error(400, "no stored data for build: {} (stored == false)".format(build_name))
        try:
            body = self.req.json_body
        except:
            return self.Error(400, "invalid target object (problem deserializing, bad JSON?)")

        target = {
            "servers": list(),
            "gitdeploys": list()
        }
        try:
            rolling_divisor = int(self.req.params['rolling_divisor']) if 'rolling_divisor' in self.req.params else 2
        except ValueError:
            return self.Error(400, "invalid rolling_divisor")
        if rolling_divisor < 1:
            return self.Error(400, "invalid rolling_divisor: {}".format(rolling_divisor))

        try:
            rolling_pause = int(self.req.params['rolling_pause']) if 'rolling_pause' in self.req.params else 15
        except ValueError:
            return self.Error(400, "invalid rolling_pause")
        if rolling_pause < 1:
            return self.Error(400, "invalid rolling_pause: {}".format(rolling_pause))

        if "environments" in body and "groups" in body:

            for c in ('environments', 'groups'):
                if not isinstance(body[c], list):
                    return self.Error(400, "{} must be a list".format(c))

            envs = self.datasvc.serversvc.GetEnvironments()
            if not self.check_against_existing(envs.keys(), body['environments']):
                return self.Error(400, "unknown environments: {}".format(self.get_unknown(envs.keys(), body['environments'])))

            existing_groups = self.datasvc.groupsvc.GetGroups(app)
            if not self.check_against_existing(existing_groups, body['groups']):
                return self.Error(400, "unknown groups: {}".format(self.get_unknown(existing_groups, body['groups'])))

            #flat list of all servers in environments
            servers_in_environments = set([s for s in [envs[e] for e in envs] for s in s])
            servers_in_groups = set([s for s in [self.datasvc.groupsvc.GetGroupServers(app, g) for g in body['groups']]
                                 for s in s])
            target['servers'] = list(servers_in_environments.intersection(servers_in_groups))
            target['gitdeploys'] = set()
            for group in body['groups']:
                group = self.datasvc.groupsvc.GetGroup(app, group)
                assert 'gitdeploys' in group
                assert isinstance(group['gitdeploys'], list)
                assert len(group['gitdeploys']) > 0
                if isinstance(group['gitdeploys'][0], list):
                    target['gitdeploys'].update([gd for sublist in group['gitdeploys'] for gd in sublist])
                else:
                    target['gitdeploys'].update(group['gitdeploys'])
            target['gitdeploys'] = list(target['gitdeploys'])

            logging.debug("DeploymentContainerView: calculated target: {}".format(target))

            for k in target:
                if target[k] < 1:
                    return self.Error(400, "no matching {} for environments ({}) and groups ({})"
                                      .format(k, body['environments'], body['groups']))

        if "servers" in body and "gitdeploys" in body:

            for c in ('servers', 'gitdeploys'):
                if not isinstance(body[c], list):
                    return self.Error(400, "{} must be a list".format(c))

            existing_servers = self.datasvc.serversvc.GetServers()
            if not self.check_against_existing(existing_servers, body['servers']):
                return self.Error(400, "unknown servers: {}".format(self.get_unknown(existing_servers, body['servers'])))

            existing_gitdeploys = self.datasvc.gitsvc.GetGitDeploys(app)
            if not self.check_against_existing(existing_gitdeploys, body['gitdeploys']):
                return self.Error(400, "unknown gitdeploys: {}".format(self.get_unknown(existing_gitdeploys,
                                                                                   body['gitdeploys'])))
            for s in body['servers']:
                target['servers'].append(s)
            for gd in body['gitdeploys']:
                target['gitdeploys'].append(gd)

            #verify that all servers have the requested gitdeploys initialized on them
            uninit_gd = dict()
            for gd in target['gitdeploys']:
                gddoc = self.datasvc.gitsvc.GetGitDeploy(app, gd)
                init_servers = set(tuple(gddoc['servers']))
                req_servers = set(tuple(target['servers']))
                if not init_servers.issuperset(req_servers):
                    uninit_gd[gd] = list(req_servers - init_servers)
            if len(uninit_gd) > 0:
                return self.Error(400, {"message": "gitdeploy not initialized on servers", "servers": uninit_gd})

        if len(target['servers']) == 0:
            return self.Error(400, "no servers specified for deployment (are you sure the gitdeploys are initialized?)")

        # do a final check to make sure all required packages are present in the build
        gddocs = [self.datasvc.gitsvc.GetGitDeploy(app, gd) for gd in target['gitdeploys']]
        for gd in gddocs:
            if gd['package'] not in build_obj.packages:
                return self.Error(400, "package {} not found in build {} (required by gitdeploy {})"
                                  .format(gd['package'], build_name, gd['name']))

        environments = body['environments'] if 'environments' in body else None
        groups = body['groups'] if 'groups' in body else None
        dpo = self.datasvc.deploysvc.NewDeployment(app, build_name, environments, groups, target['servers'], target['gitdeploys'])
        d_id = dpo['NewDeployment']['id']
        target['environments'] = environments if isinstance(environments, list) else None
        target['groups'] = groups if isinstance(groups, list) else None
        args = {
            'application': app,
            'build_name': build_name,
            'target': target,
            'rolling_divisor': rolling_divisor,
            'rolling_pause': rolling_pause,
            'deployment_id': d_id
        }
        msg = self.run_async('deploy_{}_{}'.format(app,  build_name), "deployment", args,
                             elita.deployment.deploy.run_deploy, args)
        self.datasvc.deploysvc.UpdateDeployment(app, d_id, {'status': 'running', 'job_id': msg['job_id']})
        return self.status_ok({
            'deployment': {
                'deployment_id': dpo['NewDeployment']['id'],
                'application': app,
                'build': build_name,
                'environments': environments,
                'groups': groups,
                'rolling_divisor': rolling_divisor,
                'rolling_pause': rolling_pause,
                'servers': target['servers'],
                'gitdeploys': target['gitdeploys'],
                'message': msg
            }
        })


class DeploymentView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name=context.application)
        self.set_params({"GET": [], "PUT": [], "POST": [], "DELETE": []})

    def GET(self):
        return {
            'deployment': {
                'id': self.context.name,
                'job_id': self.context.job_id,
                'created_datetime': self.get_created_datetime_text(),
                'application': self.context.application,
                'deployment': self.context.deploy if hasattr(self.context, 'deploy') else None,
                'environments': self.context.environments if hasattr(self.context, 'environments') else None,
                'groups': self.context.groups if hasattr(self.context, 'groups') else None,
                'build': self.context.build_name,
                'servers': self.context.servers if hasattr(self.context, 'servers') else None,
                'gitdeploys': self.context.gitdeploys if hasattr(self.context, 'gitdeploys') else None,
                'status': self.context.status,
                'progress': self.context.progress
            }
        }


class KeyPairContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name='_global')
        self.set_params({"GET": [], "PUT": ['name', 'key_type', 'from'], "POST": ['name', 'key_type', 'from'], "DELETE": ['name']})

    def GET(self):
        return {
            'keypairs': self.datasvc.keysvc.GetKeyPairs()
        }

    def new_keypair(self, name, private_key, public_key, attributes, key_type):
        ret = self.datasvc.keysvc.NewKeyPair(name, attributes, key_type, private_key, public_key)
        if ret['NewKeyPair']['status'] == 'ok':
            return self.status_ok({'new_keypair': {'name': name, 'key_type': key_type}})
        else:
            return self.Error(500, ret)

    def PUT(self):
        name = self.req.params['name']
        key_type = self.req.params['key_type']
        if self.req.params['from'] not in ('json', 'files'):
            return self.Error(400, "invalid 'from' type: {}; must be 'json' or 'files'".format(self.req.params['from']))
        try:
            info = self.req.json_body
        except:
            return self.Error(400, "invalid keypair info object (problem deserializing, bad JSON?)")
        attributes = info['attributes'] if 'attributes' in info else dict()
        if 'private_key' not in info:
            return self.Error(400, "private_key missing")
        else:
            private_key = info['private_key']
        if 'public_key' not in info:
            return self.Error(400, "public_key missing")
        else:
            public_key = info['public_key']
        return self.new_keypair(name, private_key, public_key, attributes, key_type)

    def POST(self):
        name = self.req.params['name']
        key_type = self.req.params['key_type']
        attributes = dict()
        logger.debug("POST keys: {}".format(self.req.POST.keys()))
        for kt in ("private_key", "public_key"):
            if kt not in self.req.POST:
                return self.Error(400, "{} not found in POST data".format(kt))
            pr_file = self.req.POST['private_key'].file
            pr_file.seek(0)
            pk_file = self.req.POST['public_key'].file
            pk_file.seek(0)
            private_key = pr_file.read(-1)
            public_key = pk_file.read(-1)

        return self.new_keypair(name, private_key, public_key, attributes, key_type)


    def DELETE(self):
        name = self.req.params['name']
        self.datasvc.keysvc.DeleteKeyPair(name)
        return self.status_ok({'delete_keypair': {'name': name}})


class KeyPairView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name='_global')
        self.set_params({"GET": [], "PUT": [], "POST": [], "DELETE": []})

    def GET(self):
        return {
            'keypair': {
                'created_datetime': self.get_created_datetime_text(),
                'name': self.datasvc.keysvc.GetKeyPair(self.context.name)
            }
        }

    def POST(self):
        try:
            info = self.req.json_body
        except:
            return self.Error(400, "invalid keypair info object (problem deserializing, bad JSON?)")
        self.datasvc.keysvc.UpdateKeyPair(self.context.name, info)
        return self.status_ok({'update_keypair': self.datasvc.keysvc.GetKeyPair(self.context.name)})

    def DELETE(self):
        self.datasvc.keysvc.DeleteKeyPair(self.context.name)
        return self.status_ok({'delete_keypair': {'name': self.context.name}})


class GitProviderContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request)
        self.set_params({"GET": [], "PUT": ['name'], "POST": ['name'], "DELETE": ['name']})

    def GET(self):
        return {
            'gitproviders': self.datasvc.gitsvc.GetGitProviders()
        }

    def PUT(self):
        name = self.req.params['name']
        try:
            info = self.req.json_body
        except:
            return self.Error(400, "invalid gitprovider info object (problem deserializing, bad JSON?)")
        if 'auth' not in info or 'type' not in info:
            return self.Error(400, "invalid gitprovider info object (valid JSON but missing one or more of: auth, type)")
        gp_type = info['type']
        auth = info['auth']
        if gp_type not in elita.deployment.gitservice.ValidRepoTypes.type_names:
            return self.Error(400, "gitprovider type not supported")
        self.datasvc.gitsvc.NewGitProvider(name, gp_type, auth)
        return self.status_ok({
            'new_gitprovider': {
                'name': name,
                'type': gp_type
            }
        })

    def POST(self):
        return self.PUT()

    def DELETE(self):
        name = self.req.params['name']
        if name in self.datasvc.gitsvc.GetGitProviders():
            self.datasvc.gitsvc.DeleteGitProvider(name)
            return self.status_ok({'delete_gitprovider': {'name': name}})
        else:
            return self.Error(400, "gitprovider '{}' not found".format(name))


class GitProviderView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request)
        self.set_params({"GET": [], "PUT": [], "POST": [], "DELETE": []})

    def GET(self):
        return {
            'gitprovider': {
                'created_datetime': self.get_created_datetime_text(),
                'name': self.context.name,
                'type': self.context.type,
            }
        }

    def POST(self):
        new_doc = dict()
        if 'type' in self.req.params:
            pass

    def DELETE(self):
        self.datasvc.gitsvc.DeleteGitProvider(self.context.name)
        return self.status_ok({
            'delete_gitprovider': {
                'name': self.context.name
            }
        })

class GitRepoContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name=context.parent)
        self.set_params({"GET": [], "PUT": ['name', 'existing', 'gitprovider', 'keypair'], "POST": ['name', 'existing'], "DELETE": ['name']})

    def GET(self):
        return {
            'application': self.context.parent,
            'gitrepos': self.datasvc.gitsvc.GetGitRepos(self.context.parent)
        }

    def PUT(self):
        existing = self.req.params['existing'] in AFFIRMATIVE_SYNONYMS
        gitprovider = self.req.params['gitprovider']
        keypair = self.req.params['keypair']
        name = self.req.params['name']

        if not existing:
            uri = None
            kp = self.datasvc.keysvc.GetKeyPair(keypair)
            if not kp:
                return self.Error(400, "unknown keypair")
            gp_doc = self.datasvc.gitsvc.GetGitProvider(gitprovider)
            if not gp_doc:
                return self.Error(400, "unkown gitprovider")
            logger.debug("GitRepoContainerView: gp_doc: {}".format(gp_doc))
            repo_callable = elita.deployment.gitservice.create_repo_callable_from_type(gp_doc['type'])
            if not repo_callable:
                return self.Error(400, "gitprovider type not supported ({})".format(gp_doc['type']))
            args = {
                'gitprovider': gp_doc,
                'name': name,
                'application': self.context.parent,
                'keypair': kp
            }
            job_data = {
                'gitprovider': gp_doc['name'],
                'name': name,
                'application': self.context.parent,
                'keypair': kp['name']
            }
            msg = self.run_async("create_repository", "async", job_data, repo_callable, args)
            ret = self.datasvc.gitsvc.NewGitRepo(self.context.parent, name, keypair, gitprovider, uri=uri)
            if ret['NewGitRepo'] != 'ok':
                return self.Error(500, {"error_NewGitRepo": ret, "message": msg})
        else:
            if 'uri' not in self.req.params:
                return self.MISSING_PARAMETER('uri')
            uri = self.req.params['uri']
            ret = self.datasvc.gitsvc.NewGitRepo(self.context.parent, name, keypair, gitprovider, uri=uri)
            if ret['NewGitRepo'] != 'ok':
                return self.Error(500, ret)
            gitrepo = self.datasvc.gitsvc.GetGitRepo(self.context.parent, name)
            args = {
                'gitrepo': gitrepo
            }
            msg = self.run_async("create_extant_gitrepo", "async", {'gitrepo': gitrepo['name']},
                                 elita.deployment.gitservice.setup_local_gitrepo_dir, args)
        return self.status_ok({
            'new_gitrepo': ret,
            'message': msg
        })

    def POST(self):
        return self.PUT()

    def DELETE(self):
        name = self.req.params['name']
        if name in self.datasvc.gitsvc.GetGitRepos():
            self.datasvc.gitsvc.DeleteGitRepo(name)
            return self.status_ok({'delete_gitrepo': {'name': name}})
        else:
            return self.Error(400, "gitrepo '{}' not found".format(name))

class GitRepoView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name=context.application)
        self.set_params({"GET": [], "PUT": [], "POST": [], "DELETE": []})

    def GET(self):
        gitrepo = self.datasvc.gitsvc.GetGitRepo(self.context.application, self.context.name)
        assert gitrepo and 'gitprovider' in gitrepo and 'auth' in gitrepo['gitprovider']
        #re-reference embedded fields
        gitrepo['gitprovider'] = gitrepo['gitprovider']['name']
        gitrepo['keypair'] = gitrepo['keypair']['name']
        return {
            'gitrepo': gitrepo
        }

    def DELETE(self):
        name = self.context.name
        app = self.context.application
        self.datasvc.gitsvc.DeleteGitRepo(app, name)
        resp = {'delete_gitrepo': {'application': app, 'name': name}}
        if 'delete' in self.req.params and self.req.params['delete'] in AFFIRMATIVE_SYNONYMS:
            gp_doc = self.datasvc.Dereference(self.context.gitprovider)
            del_callable = elita.deployment.gitservice.delete_repo_callable_from_type(gp_doc['type'])
            if not del_callable:
                return self.Error(400, "git provider type not supported ({})".format(gp_doc['type']))
            args = {
                'gitprovider': gp_doc,
                'name': self.context.name
            }
            job_data = {
                'gitprovider': gp_doc['name'],
                'name': self.context.name
            }
            msg = self.run_async("delete_repository", "async", job_data, del_callable, args)
            resp['message'] = { 'delete_repository': msg}
        return self.status_ok(resp)


class GitDeployContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name=context.parent)
        self.set_params({"GET": [], "PUT": ['name'], "POST": [], "DELETE": []})

    def GET(self):
        return {
            'application': self.context.parent,
            'gitdeploys': self.datasvc.gitsvc.GetGitDeploys(self.context.parent)
        }

    def PUT(self):
        name = self.req.params['name']
        app = self.context.parent
        try:
            location_attribs = self.req.json_body
        except:
            return self.Error(400, "invalid location/attributes object (problem deserializing, bad JSON?)")
        attribs = location_attribs['attributes'] if 'attributes' in location_attribs else {}
        options = location_attribs['options'] if 'options' in location_attribs else None
        actions = location_attribs['actions'] if 'actions' in location_attribs else None
        package = location_attribs['package'] if 'package' in location_attribs else 'master'
        if 'location' in location_attribs:
            location = location_attribs['location']
        else:
            return self.Error(400, "invalid location object (valid JSON, 'location' key not found)")
        try:
            assert 'path' in location
            assert 'gitrepo' in location
            assert 'default_branch' in location
        except AssertionError:
            return self.Error(400, "invalid location object: one or more of ('path', 'gitrepo', 'default_branch') not found")
        if location['gitrepo'] not in self.datasvc.gitsvc.GetGitRepos(app):
            return self.Error(400, "unknown gitrepo: {}".format(location['gitrepo']))
        res = self.datasvc.gitsvc.NewGitDeploy(name, app, package, options, actions, location, attribs)
        if 'error' in res:
            return self.Error(500, {"NewGitDeploy": res})
        gd = self.datasvc.gitsvc.GetGitDeploy(app, name)
        args = {'gitdeploy': gd}
        msg = self.run_async("create_gitdeploy", "async", {'gitdeploy': gd['name']},
                             elita.deployment.gitservice.create_gitdeploy, args)
        return self.status_ok({'create_gitdeploy': {'name': name, 'message': msg}})


class GitDeployView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name=context.application)
        self.set_params({"GET": [], "PUT": [], "POST": [], "DELETE": []})

    def GET(self):
        gddoc = self.datasvc.gitsvc.GetGitDeploy(self.context.application, self.context.name)
        #re-reference embedded
        gddoc['location']['gitrepo'] = gddoc['location']['gitrepo']['name']
        return {k: gddoc[k] for k in gddoc if k[0] != '_'}

    def check_servers_list(self, body):
        if "servers" not in body:
            return False, self.Error(400, "initialization requested but 'servers' not found")
        servers = body['servers']
        eservers = self.datasvc.serversvc.GetServers()
        if isinstance(servers, list):
            sset = set(eservers)
            sset_r = set(servers)
            if not sset.issuperset(sset_r):
                return False, self.Error(400, "unknown servers: {}".format(list(sset_r-sset)))
        else:
            sglob = servers
            servers = fnmatch.filter(eservers, servers)
            if len(servers) == 0:
                return False, self.Error(400, "no servers matched pattern: {}".format(sglob))
        self.servers = servers
        return True, None

    def update(self, body):
        keys = {'attributes', 'options', 'package', 'actions', 'location'}
        if "location" not in body:
            return self.Error(400, "invalid location object (valid JSON, 'location' key not found)")
        #verify no unknown/incorrect location keys
        body_keys = {k for k in body.keys()}
        if not keys.issuperset(body_keys):
            diff = body_keys - keys
            u_i = len(diff)
            word = "key"
            word += "s" if u_i > 1 else ""
            return self.Error(400, "invalid {}: {}".format(word, diff))
        self.datasvc.gitsvc.UpdateGitDeploy(self.context.application, self.context.name, body)
        gd = self.datasvc.gitsvc.GetGitDeploy(self.context.application, self.context.name)
        args = {'gitdeploy': gd}
        msg = self.run_async("create_gitdeploy", "async", {'gitdeploy': gd['name']},
                             elita.deployment.gitservice.create_gitdeploy, args)
        return self.status_ok({'update_gitdeploy': {'name': self.context.name, 'object': body, 'message': msg}})

    def initialize(self, body):
        ok, err = self.check_servers_list(body)
        if not ok:
            return err
        #we need to get the fully dereferenced doc
        gddoc = self.datasvc.gitsvc.GetGitDeploy(self.context.application, self.context.name)
        args = {
            'gitdeploy': gddoc,
            'server_list': self.servers
        }
        job_data = {
            'gitdeploy': gddoc['name'],
            'application': self.context.application,
            'servers': self.servers
        }
        msg = self.run_async('initialize_gitdeploy_servers', "async", job_data,
                             elita.deployment.gitservice.initialize_gitdeploy, args)
        return self.status_ok({
            'initialize_gitdeploy': {
                'application': self.context.application,
                'name': self.context.name,
                'servers': self.servers,
                'msg': msg
            }
        })

    def deinitialize(self, body):
        ok, err = self.check_servers_list(body)
        if not ok:
            return err
        gddoc = self.datasvc.gitsvc.GetGitDeploy(self.context.application, self.context.name)
        args = {
            'gitdeploy': gddoc, 'server_list': self.servers
        }
        job_data = {
            'gitdeploy': gddoc['name'],
            'application': self.context.application,
            'servers': self.servers
        }
        msg = self.run_async('deinitialize_gitdeploy_servers', "async", job_data,
                             elita.deployment.gitservice.deinitialize_gitdeploy, args)
        return self.status_ok({
            'deinitialize_gitdeploy': {
                'application': self.context.application,
                'name': self.context.name,
                'servers': self.servers,
                'msg': msg
            }
        })

    def POST(self):
        try:
            body = self.req.json_body
        except:
            return self.Error(400, "invalid body object (problem deserializing, bad JSON?)")
        if "initialize" in self.req.params and self.req.params['initialize'] in AFFIRMATIVE_SYNONYMS:
            return self.initialize(body)
        elif "deinitialize" in self.req.params and self.req.params['deinitialize'] in AFFIRMATIVE_SYNONYMS:
            return self.deinitialize(body)
        else:
            return self.update(body)

    def DELETE(self):
        gddoc = self.datasvc.gitsvc.GetGitDeploy(self.context.application, self.context.name)
        args = {'gitdeploy': gddoc}
        groups_deleted = False
        delete_groups = list()
        if "delete_groups" in self.req.params and self.req.params['delete_groups'] in AFFIRMATIVE_SYNONYMS:
            groups = self.datasvc.appsvc.GetGroups(self.context.application)
            for g in groups:
                group = self.datasvc.appsvc.GetGroup(self.context.application, g)
                if self.context.name in group['gitdeploys']:
                    delete_groups.append(g)
            for g in delete_groups:
                logger.debug("deleting group: {} (application: {})".format(g, self.context.application))
                self.datasvc.appsvc.DeleteGroup(self.context.application, g)
            groups_deleted = True

        msg = self.run_async('remove_deinitialize_gitdeploy', "async", {'gitdeploy': gddoc['name']},
                             elita.deployment.gitservice.remove_and_deinitialize_gitdeploy, args)
        status_msg = {
            'delete_deinitialize_gitdeploy': {
                'application': gddoc['application'],
                'name': gddoc['name'],
                'servers': gddoc['servers'] if 'servers' in gddoc else list(),
                'msg': msg
            }
        }
        if groups_deleted:
            status_msg['delete_deinitialize_gitdeploy']['groups_deleted'] = delete_groups
        return self.status_ok(status_msg)


class GlobalContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, permissionless=True)

    def GET(self):
        return {"global": self.datasvc.GetGlobalKeys()}


class GroupContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name=context.parent)
        self.set_params({"GET": [], "PUT": ["name", "rolling_deploy"], "POST": [], "DELETE": ["name"]})

    def GET(self):
        return {
            'application': self.context.parent,
            'groups': self.datasvc.groupsvc.GetGroups(self.context.parent)
        }

    def PUT(self):
        name = self.req.params['name']
        rolling_deploy = self.req.params['rolling_deploy'] in AFFIRMATIVE_SYNONYMS
        try:
            gitdeploys = self.req.json_body
        except:
            return self.Error(400, "invalid gitdeploys object (problem deserializing, bad JSON?)")
        if "gitdeploys" not in gitdeploys.keys():
            return self.Error(400, "JSON missing key 'gitdeploys'")
        if not isinstance(gitdeploys['gitdeploys'], list):
            return self.Error(400, "gitdeploys must be list")
        description = gitdeploys['description'] if 'description' in gitdeploys else ""
        attributes = gitdeploys['attributes'] if 'attributes' in gitdeploys else dict()

        existing_gitdeploys = set(self.datasvc.gitsvc.GetGitDeploys(self.context.parent))
        submitted_gitdeploys = set([gd for sublist in gitdeploys['gitdeploys'] for gd in sublist]) \
            if isinstance(gitdeploys['gitdeploys'][0], list) else set(gitdeploys['gitdeploys'])
        if not existing_gitdeploys.issuperset(submitted_gitdeploys):
            return self.Error(400, "unknown gitdeploys: {}".format(list(submitted_gitdeploys - existing_gitdeploys)))

        self.datasvc.groupsvc.NewGroup(self.context.parent, name, gitdeploys['gitdeploys'], description=description,
                                       attributes=attributes, rolling_deploy=rolling_deploy)
        return self.status_ok({
            "New_Group": {
                "name": name,
                "description": description,
                "application": self.context.parent,
                "attributes": attributes,
                "rolling_deploy": rolling_deploy,
                "gitdeploys": gitdeploys['gitdeploys']
            }
        })

    def DELETE(self):
        name = self.req.params['name']
        if name not in self.datasvc.groupsvc.GetGroups(self.context.parent):
            return self.Error(400, "unknown group: {}".format(name))
        self.datasvc.groupsvc.DeleteGroup(self.context.parent, name)
        return self.status_ok({
            "Delete_Group": {
                "name": name,
                "application": self.context.parent
            }
        })

class GroupView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name=context.application)

    def GET(self):
        environments = self.req.params['environments'].split(' ') \
            if ('environments' in self.req.params and len(self.req.params['environments']) > 0) else None
        if environments:
            existing_groups = self.datasvc.serversvc.GetEnvironments()
            if not self.check_against_existing(existing_groups, environments):
                return self.Error(400, "unknown environments: {}".format(self.get_unknown(existing_groups, environments)))
        return {
            'created': self.get_created_datetime_text(),
            'application': self.context.application,
            'description': self.context.description,
            'name': self.context.name,
            'attributes': self.context.attributes,
            'rolling_deploy': self.context.rolling_deploy,
            'gitdeploys': self.context.gitdeploys,
            'servers': self.datasvc.groupsvc.GetGroupServers(self.context.application, self.context.name,
                                                             environments=environments),
            'environments': environments if environments else "(any)"
        }

    def DELETE(self):
        self.datasvc.groupsvc.DeleteGroup(self.context.application, self.context.name)
        return self.status_ok({
            "Delete_Group": {
                "name": self.context.name,
                "application": self.context.application
            }
        })

class UserContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request)
        self.set_params({"GET": [], "PUT": ["username", "password"],
                         "POST": ["username", "password"], "DELETE": ["username"]})

    def PUT(self):
        name = self.req.params['username']
        pw = self.req.params['password']
        try:
            perms_attribs = self.req.json_body
        except:
            return self.Error(400, "invalid user attributes object (problem deserializing, bad JSON?)")
        if "permissions" in perms_attribs:
            perms = perms_attribs['permissions']
            attribs = perms_attribs['attributes'] if 'attributes' in perms_attribs else dict()
            if auth.ValidatePermissionsObject(perms).run():
                self.datasvc.usersvc.NewUser(name, pw, perms, attribs)
                return self.status_ok({"user_created": {"username": name, "password": "(hidden)",
                                                        "permissions": perms, "attributes": attribs}})
            else:
                return self.Error(400, "invalid permissions object (valid JSON but semantically incorrect)")
        else:
            return self.Error(400, "invalid user attributes object (missing permissions)")

    def GET(self):
        return {"users": self.datasvc.usersvc.GetUsers()}

    def POST(self):
        return self.PUT()

    def DELETE(self):
        name = self.req.params['username']
        if name in self.datasvc.usersvc.GetUsers():
            self.datasvc.usersvc.DeleteUser(name)
            return self.status_ok({"user_deleted": {"username": name}})
        else:
            return self.Error(400, "unknown user")

class JobView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, permissionless=True)  # job_id is the secret
        self.set_params({"GET": [], "PUT": [], "POST": [], "DELETE": []})

    def GET(self):
        ret = {
            'job_id': str(self.context.job_id),
            'name': self.context.name,
            'job_type': self.context.job_type,
            'data': self.context.data,
            'created_datetime': self.get_created_datetime_text(),
            'status': self.context.status,
        }
        if self.context.completed_datetime is not None:
            ret['completed_datetime'] = self.context.completed_datetime.isoformat(' ')
        if self.context.duration_in_seconds is not None:
            ret['duration_in_seconds'] = self.context.duration_in_seconds
        if 'results' in self.req.params:
            if self.req.params['results'] in AFFIRMATIVE_SYNONYMS:
                ret['results'] = self.datasvc.jobsvc.GetJobData(self.context.job_id)
        return ret

class JobContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request)
        self.set_params({"GET": ["active"], "PUT": [], "POST": [], "DELETE": []})

    def GET(self):
        active = self.req.params['active'] in AFFIRMATIVE_SYNONYMS
        return {"jobs": {"active": active, "job_ids": self.datasvc.jobsvc.GetJobs(active=active)}}


class UserView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, allow_pw_auth=True)  # allow both pw and auth_token auth
        self.set_params({"GET": [], "PUT": [], "POST": [], "DELETE": []})


    def status_ok_with_token(self, username):
        return self.status_ok({"username": username, "permissions": self.context.permissions,
                               "attributes": self.context.attributes,
                               "auth_token": self.datasvc.usersvc.GetUserTokens(username)})

    def change_password(self, new_pw):
        self.context.change_password(new_pw)
        self.datasvc.usersvc.SaveUser(self.context)

    def change_attributes(self, attribs):
        self.context.attributes = attribs
        self.datasvc.usersvc.SaveUser(self.context)

    def change_permissions(self, perms):
        self.context.permissions = perms
        self.datasvc.usersvc.SaveUser(self.context)

    def GET(self):
        # ugly. we want to support both password auth and valid tokens, but only tokens from the same user or with the
        # '_global' permission.
        if 'password' in self.req.params:
            if not self.context.validate_password(self.req.params['password']):
                return self.Error(403, "incorrect password")
        elif ('auth_token' not in self.req.params) and ('Auth-Token' not in self.req.headers):
            return self.Error(403, "password or auth token required")
        else:
            if 'auth_token' in self.req.params and len(self.req.params.getall('auth_token')) > 1:  # multiple auth tokens
                return self.Error(403, "incorrect authentication")
            token = self.req.params.getall('auth_token')[0] if 'auth_token' in self.req.params else self.req.headers['Auth-Token']
            authsvc = auth.UserPermissions(self.datasvc.usersvc, token, datasvc=self.datasvc)
            allowed_apps = authsvc.get_allowed_apps()
            global_read = False
            for k in allowed_apps:
                if 'read' in k:
                    if '_global' in allowed_apps[k]:
                        global_read = True
            if not authsvc.valid_token or authsvc.username != self.context.username or not global_read:
                logging.debug("valid_token: {}".format(authsvc.valid_token))
                logging.debug("usernames: {}, {}".format(authsvc.username, self.context.username))
                logging.debug("allowed_apps: {}".format(authsvc.get_allowed_apps()))
                return self.Error(403, "bad token")
        name = self.context.username
        if len(self.datasvc.usersvc.GetUserTokens(name)) == 0:
            self.datasvc.usersvc.NewToken(name)
        return self.status_ok_with_token(name)

    def POST(self):
        if 'password' in self.req.params:
            if not self.context.validate_password(self.req.params['password']):
                return self.Error(403, "incorrect password")
        update = self.req.params['update'] if 'update' in self.req.params else 'token'
        if update == "token":
            self.datasvc.usersvc.NewToken(self.context.name)
            return self.status_ok_with_token(self.context.name)
        elif update == "password":
            if "new_password" in self.req.params:
                self.change_password(self.req.params['new_password'])
                return self.status_ok("password changed")
            else:
                return self.Error(400, "parameter new_password required")
        elif update == "attributes":
            try:
                attribs = self.req.json_body
            except:
                return self.Error(400, "problem deserializing attributes object")
            self.change_attributes(attribs)
            return self.status_ok({"new_attributes": attribs})
        elif update == "permissions":
            try:
                perms = self.req.json_body
            except:
                return self.Error(400, "problem deserializing permissions object (bad JSON?)")
            perms = perms['permissions'] if 'permissions' in perms else perms
            if auth.ValidatePermissionsObject(perms).run():
                self.change_permissions(perms)
                return self.status_ok({"new_permissions": perms})
            else:
                return self.Error(400, "invalid permissions object (valid JSON but semantically incorrect)")
        else:
            return self.Error(400, "incorrect request type '{}'".format(self.req.params['request']))

class UserPermissionsView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, allow_pw_auth=True)
        self.set_params({"GET": [], "PUT": [], "POST": [], "DELETE": []})

    def compute_app_perms(self):
        return auth.UserPermissions(self.datasvc.usersvc, None, datasvc=self.datasvc).get_allowed_apps(self.context
                                                                                                       .username)

    def compute_action_perms(self):
        return auth.UserPermissions(self.datasvc.usersvc, None, datasvc=self.datasvc).get_allowed_actions(self
                                                                                                          .context
                                                                                                          .username)

    def compute_server_perms(self):
        return auth.UserPermissions(self.datasvc.usersvc, None, datasvc=self.datasvc).get_allowed_servers(self
                                                                                                          .context
                                                                                                          .username)

    def GET(self):
        return {
            'username': self.context.username,
            'applications': self.compute_app_perms(),
            'actions': self.compute_action_perms(),
            'servers': self.compute_server_perms()
        }


class TokenContainerView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, app_name='_global')
        self.set_params({"GET": ["username", "password"], "PUT": [], "POST": [], "DELETE": ["token"]})

    def GET(self):
        username = self.req.params['username']
        pw = self.req.params['password']
        if auth.UserPermissions(self.datasvc.usersvc, None).validate_pw(username, pw):
            tokens = self.datasvc.usersvc.GetUserTokens(username)
            if len(tokens) > 0:
                return self.status_ok({"username": username, "token": tokens})
            else:
                return self.Error(400, "no token found for '{}'".format(username))
        else:
            return self.Error(403, "incorrect password")

    def DELETE(self):
        token = self.req.params['token']
        if token in self.datasvc.usersvc.GetAllTokens():
            username = self.datasvc.usersvc.GetUserFromToken(token)
            self.datasvc.usersvc.DeleteToken(token)
            return self.status_ok({"token_deleted": {"username": username, "token": token}})
        else:
            return self.Error(400, "unknown token")

class TokenView(GenericView):
    def __init__(self, context, request):
        #permissionless b/c the token is the secret. no need to supply token in URL and auth_token param
        GenericView.__init__(self, context, request, permissionless=True)

    def GET(self):
        return {"token": self.context.token, "created": self.get_created_datetime_text(),
                "username": self.context.username}

    def DELETE(self):
        token = self.context.token
        user = self.context.username
        self.datasvc.usersvc.DeleteToken(token)
        return self.status_ok({"token_deleted": {"username": user, "token": token}})


#this is the default 'root' view that gets called for every request
@view_config(name="", renderer='json')
def Action(context, request):
    logger.debug("REQUEST: url: {}".format(request.url))
    logger.debug("REQUEST: context: {}".format(context.__class__.__name__))
    logger.debug("REQUEST: method: {}".format(request.method))
    logger.debug("REQUEST: params: {}".format(request.params))

    pp = pprint.PrettyPrinter(indent=4)
    if context.__class__.__name__ == 'Action':
        logger.debug("REQUEST: action")
        mobj = context
        cname = "Action"
    else:
        logger.debug("REQUEST: context.doc: {}".format(pp.pformat(context.doc)))
        cname = context.doc['_class']
        try:
            mobj = models.__dict__[cname](context.doc)
        except elita_exceptions.SaltServerNotAccessible:
            return {
                'server': context.doc['name'],
                'error': "Server object not accessible via salt"
            }

    logger.debug("Model class: {}".format(cname))

    view_class = globals()[cname + "View"]

    return view_class(mobj, request).__call__()

class RootView(GenericView):
    def __init__(self, context, request):
        GenericView.__init__(self, context, request, permissionless=True)

    def GET(self):
        return About(self.req)

    def POST(self):
        return self.GET()
    def PUT(self):
        return self.GET()
    def DELETE(self):
        return self.GET()

import pkg_resources
@view_config(name="about", renderer='json')
def About(request):
    apps = request.datasvc.appsvc.GetApplications()
    appstats = {
        a: {
            'builds': len(request.datasvc.buildsvc.GetBuilds(a)),
            'gitrepos': len(request.datasvc.gitsvc.GetGitRepos(a)),
            'gitdeploys': len(request.datasvc.gitsvc.GetGitDeploys(a))
        } for a in apps}
    return {
        'about': {
            'name': 'elita',
            'version': pkg_resources.require("elita")[0].version,
            'tagline': "You Know, for DevOps",
            'hostname': socket.getfqdn()
        },
        'stats': {
            'applications': appstats,
            'servers': len(request.datasvc.serversvc.GetServers()),
            'users': len(request.datasvc.usersvc.GetUsers())
        }
    }







__author__ = 'bkeroack'

import sys
import logging
import traceback
import time
import billiard
import itertools
import pprint
pp = pprint.PrettyPrinter(indent=4)

import elita.util
import elita.util.type_check
import gitservice
import salt_control
from elita.actions.action import regen_datasvc

class FatalDeploymentError(Exception):
    pass

#async callable
def run_deploy(datasvc, application, build_name, target, rolling_divisor, rolling_pause, deployment_id):
    '''
    Asynchronous entry point for deployments
    '''

    # normally there's a higher level try/except block for all async actions
    # we want to make sure the error is saved in the deployment object as well, not just the job
    # so we duplicate the functionality here
    try:
        if target['groups']:
            logging.debug("run_deploy: Doing rolling deployment")
            dc = DeployController(datasvc, deployment_id)
            rdc = RollingDeployController(datasvc, dc, deployment_id)
            ret = rdc.run(application, build_name, target, rolling_divisor, rolling_pause)
        else:
            logging.debug("run_deploy: Doing manual deployment")
            dc = DeployController(datasvc, deployment_id)
            ret, data = dc.run(application, build_name, target['servers'], target['gitdeploys'])
    except:
        exc_type, exc_obj, tb = sys.exc_info()
        f_exc = traceback.format_exception(exc_type, exc_obj, tb)
        results = {
            "error": "unhandled exception during callable!",
            "exception": f_exc
        }
        logging.debug("run_deploy: EXCEPTION: {}".format(f_exc))
        datasvc.deploysvc.UpdateDeployment(application, deployment_id, {"status": "error"})
        return {"deploy_status": "error", "details": results}
    datasvc.deploysvc.CompleteDeployment(application, deployment_id)
    datasvc.deploysvc.UpdateDeployment(application, deployment_id, {"status": "complete" if ret else "error"})
    return {"deploy_status": "done" if ret else "error"}


class BatchCompute:
    '''
        Given a list of application groups that require rolling deployment and an (optional) list that do not,
        compute the optimal batches of server/gitdeploy pairs. All non-rolling groups are added to the first batch.
        Splitting algorithm is tolerant of outrageously large rolling_divisors.

        Written in a functional style to facilitate testing.
    '''

    @staticmethod
    def add_nonrolling_groups(batches, nonrolling_docs):
        '''
        Add servers and gitdeploys from nonrolling groups to the first batch
        Not written in a functional style because that was totally unreadable
        '''
        if nonrolling_docs and len(nonrolling_docs) > 0:
            assert all(map(lambda x: 'servers' in x and 'gitdeploys' in x, batches)) or not batches
            assert all(map(lambda x: 'servers' in x[1] and 'gitdeploys' in x[1], nonrolling_docs.iteritems()))
            non_rolling_batches = list()
            for g in nonrolling_docs:
                servers = nonrolling_docs[g]['servers']
                gitdeploys = nonrolling_docs[g]['gitdeploys']
                if isinstance(nonrolling_docs[g]['gitdeploys'][0], list):
                    for i, gdb in enumerate(nonrolling_docs[g]['gitdeploys']):
                        if i > len(non_rolling_batches)-1:
                            non_rolling_batches.append({'gitdeploys': gdb, 'servers': servers})
                        else:
                            non_rolling_batches[i]['servers'] = list(set(servers + non_rolling_batches[i]['servers']))
                            non_rolling_batches[i]['gitdeploys'] = list(set(gdb + non_rolling_batches[i]['gitdeploys']))
                else:
                    non_rolling_batches.append({'gitdeploys': gitdeploys, 'servers': servers})
            for i, nrb in enumerate(non_rolling_batches):
                if i > len(batches)-1:
                    batches.append(nrb)
                else:
                    batches[i]['servers'] = list(set(nrb['servers'] + batches[i]['servers']))
                    batches[i]['gitdeploys'] = list(set(nrb['gitdeploys'] + batches[i]['gitdeploys']))
        return batches

    @staticmethod
    def dedupe_batches(batches):
        '''
        Dedupe servers and gitdeploys list in the combined batches list:
        [
            { "servers": [ "server1", "server1", ...], "gitdeploys": [ "gd1", "gd1", ...] },  #batch 0 (all groups)
            { "servers": [ "server1", "server1", ...], "gitdeploys": [ "gd1", "gd1", ...] },  #batch 1 (all groups)
            ...
        ]
        '''
        assert len(batches) > 0
        assert all(map(lambda x: 'servers' in x and 'gitdeploys' in x, batches))
        return map(lambda x: {"servers": list(set(x['servers'])), "gitdeploys": list(set(elita.util.flatten_list(x['gitdeploys'])))}, batches)

    @staticmethod
    def reduce_group_batches(accumulated, update):
        assert 'servers' in accumulated and 'servers' in update
        assert 'gitdeploys' in accumulated and 'gitdeploys' in update
        return {
            "servers": accumulated['servers'] + update['servers'],
            "gitdeploys": accumulated['gitdeploys'] + update['gitdeploys']
        }

    @staticmethod
    def coalesce_batches(batches):
        '''
        Combine the big list of batches into a single list.

        Function is passed a list of lists:
        [
            [ { "servers": [...], "gitdeploys": [...] }, ... ],     # batches 0-n for group A
            [ { "servers": [...], "gitdeploys": [...] }, ... ],     # batches 0-n for broup B
            ...
        ]

        Each nested list represents the computed batches for an individual group. All nested lists are expected to be
        the same length.
        '''
        if not batches:
            return list()

        return map(
            lambda batch_aggregate: reduce(
                lambda acc, upd:
                {
                    'servers': acc['servers'] + upd['servers'],
                    'gitdeploys': acc['gitdeploys'] + upd['gitdeploys']
                }, batch_aggregate
            ), itertools.izip_longest(*batches, fillvalue={"servers": [], "gitdeploys": []}))

    @staticmethod
    def compute_group_batches(divisor, group):
        '''
        Compute batches for group.
        Group is iteritems() result from group dict. group[0] is key (name), group[1] is dict of servers/gitdeploys

        return list of dicts: [ { 'servers': [...], 'gitdeploys': [...] }, ... ]
        '''
        assert len(group) == 2
        assert 'servers' in group[1]
        assert 'gitdeploys' in group[1]

        servers = group[1]['servers']
        gitdeploys = group[1]['gitdeploys']
        server_batches = elita.util.split_seq(servers, divisor)
        gd_multiplier = len(server_batches)  # gitdeploy_batches multipler
        if isinstance(gitdeploys[0], list):
            # duplicate all server batches by the length of the gitdeploy list-of-lists
            server_batches = [x for item in server_batches for x in itertools.repeat(item, len(gitdeploys))]
            gitdeploy_batches = list(gitdeploys) * gd_multiplier
        else:
            gitdeploy_batches = [gitdeploys] * gd_multiplier
        assert len(gitdeploy_batches) == len(server_batches)
        return [{'servers': sb, 'gitdeploys': gd} for sb, gd in zip(server_batches, gitdeploy_batches)]

    @staticmethod
    def compute_rolling_batches(divisor, rolling_group_docs, nonrolling_group_docs):
        assert isinstance(divisor, int)
        assert elita.util.type_check.is_optional_dict(rolling_group_docs)
        assert not rolling_group_docs or all(map(lambda x: 'servers' in x[1] and 'gitdeploys' in x[1], rolling_group_docs.iteritems()))
        return BatchCompute.dedupe_batches(
            BatchCompute.add_nonrolling_groups(
                BatchCompute.coalesce_batches(
                    map(lambda x: BatchCompute.compute_group_batches(divisor, x), rolling_group_docs.iteritems() if rolling_group_docs else tuple())
                ), nonrolling_group_docs
            )
        )


class RollingDeployController:
    '''
    Break deployment up into server/gitdeploy batches, then invoke DeployController with each batch sequentially
    '''
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, datasvc, deploy_controller, deployment_id):
        '''
        @type datasvc: models.DataService
        '''
        self.datasvc = datasvc
        self.dc = deploy_controller
        self.deployment_id = deployment_id

    def get_nonrolling_groups(self, rolling_groups, all_groups):
        return list(set(all_groups) - set(rolling_groups))

    def compute_batches(self, rolling_group_docs, nonrolling_group_docs, rolling_divisor):
        return BatchCompute.compute_rolling_batches(rolling_divisor, rolling_group_docs, nonrolling_group_docs)

    def run(self, application, build_name, target, rolling_divisor, rolling_pause, parallel=True):
        '''
        Run rolling deployment. This should be called iff the deployment is called via groups/environments
        '''

        groups = target['groups']
        rolling_groups = [g for g in groups if self.datasvc.groupsvc.GetGroup(application, g)['rolling_deploy']]
        rolling_group_docs = {g: self.datasvc.groupsvc.GetGroup(application, g) for g in rolling_groups}
        nonrolling_groups = self.get_nonrolling_groups(rolling_groups, groups)
        nonrolling_group_docs = {g: self.datasvc.groupsvc.GetGroup(application, g) for g in nonrolling_groups}

        gd_docs = [self.datasvc.gitsvc.GetGitDeploy(application, gd) for gd in target['gitdeploys']]
        gitrepos = [gd['location']['gitrepo']['name'] for gd in gd_docs]

        batches = self.compute_batches(rolling_group_docs, nonrolling_group_docs, rolling_divisor)

        logging.debug("computed batches: {}".format(batches))

        self.datasvc.deploysvc.InitializeDeploymentPlan(application, self.deployment_id, batches, gitrepos)

        self.datasvc.jobsvc.NewJobData({
            "RollingDeployment": {
                "batches": len(batches),
                "batch_data": batches
            }
        })

        for i, b in enumerate(batches):
            logging.debug("doing DeployController.run: deploy_gds: {}".format(b['gitdeploys']))
            ok, results = self.dc.run(application, build_name, b['servers'], b['gitdeploys'], force=i > 0,
                                      parallel=parallel, batch_number=i)
            if not ok:
                self.datasvc.jobsvc.NewJobData({"RollingDeployment": "error"})
                return False

            deploy_doc = self.datasvc.deploysvc.GetDeployment(application, self.deployment_id)
            assert deploy_doc
            if deploy_doc['status'] == 'error':
                self.datasvc.jobsvc.NewJobData({"message": "detected failed deployment so aborting further batches"})
                return False

            if i != (len(batches)-1):
                msg = "pausing for {} seconds between batches".format(rolling_pause)
                self.datasvc.jobsvc.NewJobData({"RollingDeployment": msg})
                logging.debug("RollingDeployController: {}".format(msg))
                time.sleep(rolling_pause)

        return True

def determine_deployabe_servers(all_gd_servers, specified_servers):
    return list(set(all_gd_servers).intersection(set(specified_servers)))

def _threadsafe_process_gitdeploy(gddoc, build_doc, servers, queue, settings, job_id, deployment_id):
    '''
    Threadsafe function for processing a single gitdeploy during a deployment.
    Creates own instance of datasvc, etc.
    Pushes changed gitdeploys to a shared queue
    '''

    package = gddoc['package']
    package_doc = build_doc['packages'][package]

    client, datasvc = regen_datasvc(settings, job_id)
    gdm = gitservice.GitDeployManager(gddoc, datasvc)
    gitrepo_name = gddoc['location']['gitrepo']['name']

    datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                              progress=10,
                                              step='Checking out default branch')

    try:
        res = gdm.checkout_default_branch()
    except:
        exc_msg = str(sys.exc_info()[1]).split('\n')
        exc_msg.insert(0, "ERROR: checkout_default_branch")
        datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                                  step=exc_msg)
        return

    logging.debug("_threadsafe_process_gitdeploy: git checkout output: {}".format(str(res)))

    if gdm.last_build == build_doc['build_name']:
        datasvc.jobsvc.NewJobData({"ProcessGitdeploys": {gitrepo_name: "already processed"}})
        datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                              progress=100,
                                              step='Complete (already processed)')
    else:
        datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                              progress=25,
                                              step='Decompressing package to repository')

        datasvc.jobsvc.NewJobData({"ProcessGitdeploys": {gddoc['name']: "processing"}})

        try:
            gdm.decompress_to_repo(package_doc)
        except:
            exc_msg = str(sys.exc_info()[1]).split('\n')
            exc_msg.insert(0, "ERROR: decompress_to_repo")
            datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                                      step=exc_msg)
            return

        datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                              progress=50,
                                              step='Checking for changes')

        logging.debug("_threadsafe_process_gitdeploy: Checking for changes")
        try:
            res = gdm.check_repo_status()
        except:
            exc_msg = str(sys.exc_info()[1]).split('\n')
            exc_msg.insert(0, "ERROR: check_repo_status")
            datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                                      step=exc_msg)
            return
        logging.debug("_threadsafe_process_gitdeploy: git status results: {}".format(str(res)))

        if "nothing to commit" in res:
            datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                              progress=100,
                                              step='Complete (no changes found)')
            datasvc.jobsvc.NewJobData({"ProcessGitdeploys": {gddoc['name']: "no changes"}})
        else:
            datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                              progress=60,
                                              step='Adding changes to repository')
            datasvc.jobsvc.NewJobData({"ProcessGitdeploys": {gddoc['name']: "adding to repository"}})
            try:
                res = gdm.add_files_to_repo()
            except:
                exc_msg = str(sys.exc_info()[1]).split('\n')
                exc_msg.insert(0, "ERROR: add_files_to_repo")
                datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                                          step=exc_msg)
                return
            logging.debug("_threadsafe_process_gitdeploy: git add result: {}".format(str(res)))

            datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                              progress=70,
                                              step='Committing changes to repository')
            datasvc.jobsvc.NewJobData({"ProcessGitdeploys": {gddoc['name']: "committing"}})
            try:
                res = gdm.commit_to_repo(build_doc['build_name'])
            except:
                exc_msg = str(sys.exc_info()[1]).split('\n')
                exc_msg.insert(0, "ERROR: commit_to_repo")
                datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                                          step=exc_msg)
                return
            logging.debug("_threadsafe_process_gitdeploy: git commit result: {}".format(str(res)))

            datasvc.jobsvc.NewJobData({"ProcessGitdeploys": {gddoc['name']: "checking diff"}})
            try:
                res = gdm.inspect_latest_diff()
            except:
                exc_msg = str(sys.exc_info()[1]).split('\n')
                exc_msg.insert(0, "ERROR: inspect_latest_diff")
                datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                                          step=exc_msg)
                return
            logging.debug("_threadsafe_process_gitdeploy: inspect diff result: {}".format(str(res)))

            # change to a list of dicts without filenames as keys to keep mongo happy
            changed_files = [{
                'filename': k,
                'deletions': res[k]['deletions'],
                'lines': res[k]['lines'],
                'insertions': res[k]['insertions']
            } for k in res]

            datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                              changed_files=changed_files)

            datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                              progress=90,
                                              step='Pushing changes to gitprovider')
            datasvc.jobsvc.NewJobData({"ProcessGitdeploys": {gddoc['name']: "pushing"}})
            try:
                res = gdm.push_repo()
            except:
                exc_msg = str(sys.exc_info()[1]).split('\n')
                exc_msg.insert(0, "ERROR: push_repo")
                datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                                          step=exc_msg)
                return
            logging.debug("_threadsafe_process_gitdeploy: git push result: {}".format(str(res)))
            datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                              progress=100,
                                              step='Complete')
        try:
            gdm.update_repo_last_build(build_doc['build_name'])
        except:
            exc_msg = str(sys.exc_info()[1]).split('\n')
            exc_msg.insert(0, "ERROR: update_repo_last_build")
            datasvc.deploysvc.UpdateDeployment_Phase1(gddoc['application'], deployment_id, gitrepo_name,
                                                      step=exc_msg)
            return
    # in the event that the gitrepo hasn't changed, but the gitdeploy indicates that we haven't successfully
    # deployed to all servers, we want to force git pull
    # this can happen if multiple gitdeploys share the same gitrepo

    #always pull for now
    queue.put_nowait({gddoc['name']: determine_deployabe_servers(gddoc['servers'], servers)})

def _threadsafe_pull_callback(results, tag, **kwargs):
    '''
    Passed to run_slses_async and is used to provide realtime updates to users polling the deploy job object
    '''
    try:
        assert all([arg in kwargs for arg in ('datasvc', 'application', 'deployment_id', 'batch_number', 'gitdeploy')])
    except AssertionError:
        #can't log anything to the job object because we may not have a valid DataService instance
        logging.error("***************** _threadsafe_pull_callback: AssertionError: incorrect kwargs ****************")
        return

    datasvc = kwargs['datasvc']
    app = kwargs['application']
    deployment_id = kwargs['deployment_id']
    batch_number = kwargs['batch_number']
    gitdeploy = kwargs['gitdeploy']
    try:
        datasvc.jobsvc.NewJobData({"DeployServers": {"results": results, "tag": tag}})
        for r in results:
            #callback results always have a 'ret' key but underneath it may be a simple string or a big complex object
            #for state call results. We have to unpack the state results if necessary
            this_result = results[r]['ret']
            datasvc.jobsvc.NewJobData(this_result)
            if elita.util.type_check.is_dictlike(this_result):  # state result
                for state_res in this_result:
                    if "result" in this_result[state_res]:
                        state_comment = this_result[state_res]['comment'] if 'comment' in this_result[state_res] else state_res
                        stdout = this_result[state_res]["changes"]["stdout"] if 'changes' in this_result[state_res] and 'stdout' in this_result[state_res]['changes'] else "(none)"
                        stderr = this_result[state_res]["changes"]["stderr"] if 'changes' in this_result[state_res] and 'stderr' in this_result[state_res]['changes'] else "(none)"
                        if not this_result[state_res]["result"]:    #error
                            logging.debug("_threadsafe_pull_callback: got error result ({}; {})".format(gitdeploy, r))
                            datasvc.jobsvc.NewJobData({'status': 'error', 'message': 'failing deployment due to detected error'})
                            datasvc.deploysvc.UpdateDeployment_Phase2(app, deployment_id, gitdeploy, [r], batch_number,
                                                                      state="FAILURE: {}; stderr: {}; stdout: {}".format(state_comment, stderr, stdout))
                            datasvc.deploysvc.FailDeployment(app, deployment_id)
                        else:
                            logging.debug("_threadsafe_pull_callback: got successful result ({}; {}): {}".format(gitdeploy, r, state_comment))
                            datasvc.deploysvc.UpdateDeployment_Phase2(app, deployment_id, gitdeploy, [r], batch_number,
                                                          state=state_comment,
                                                          progress=66)
            else:   # simple result
                logging.debug("_threadsafe_pull_callback: got simple return instead of results ({}; {})".format(gitdeploy, r))
                datasvc.deploysvc.UpdateDeployment_Phase2(app, deployment_id, gitdeploy, [r], batch_number,
                                                          state="simple return: {}".format(results[r]['ret']))
    except:
        exc_type, exc_obj, tb = sys.exc_info()
        datasvc.jobsvc.NewJobData({"_threadsafe_pull_callback EXCEPTION": traceback.format_exception(exc_type, exc_obj, tb)})
        datasvc.deploysvc.FailDeployment(app, deployment_id)

def _threadsafe_pull_gitdeploy(application, gitdeploy_struct, queue, settings, job_id, deployment_id, batch_number):
    '''
    Thread-safe way of performing a deployment SLS call for one specific gitdeploy on a group of servers
    gitdeploy_struct: { "gitdeploy_name": [ list_of_servers_to_deploy_to ] }
    '''
    # Wrap in a big try/except so we can log any failures in phase2 progress and fail the deployment
    try:
        assert settings
        assert job_id
        assert gitdeploy_struct
        client, datasvc = regen_datasvc(settings, job_id)
        gd_name = gitdeploy_struct.keys()[0]
        servers = gitdeploy_struct[gd_name]
    except:
        exc_msg = str(sys.exc_info()[1]).split('\n')
        logging.error("************* _threadsafe_pull_gitdeploy: preamble: {} *********************".format(exc_msg))
        return
    try:
        assert application
        assert queue
        assert deployment_id
        assert all([elita.util.type_check.is_string(gd) for gd in gitdeploy_struct])
        assert all([elita.util.type_check.is_seq(gitdeploy_struct[gd]) for gd in gitdeploy_struct])
        assert isinstance(batch_number, int) and batch_number >= 0
        sc = salt_control.SaltController(datasvc)
        rc = salt_control.RemoteCommands(sc)
        assert len(gitdeploy_struct) == 1

        datasvc.deploysvc.UpdateDeployment_Phase2(application, deployment_id, gd_name, servers, batch_number,
                                                  progress=10,
                                                  state="Beginning deployment")

        #until salt Helium is released, we can only execute an SLS *file* as opposed to a single module call
        sls_map = {sc.get_gitdeploy_entry_name(application, gd_name): servers}
        if len(servers) == 0:
            datasvc.jobsvc.NewJobData({"DeployServers": {gd_name: "no servers"}})
            return True

        #clear uncommitted changes on targets
        datasvc.deploysvc.UpdateDeployment_Phase2(application, deployment_id, gd_name, servers, batch_number,
                                                  progress=33,
                                                  state="Clearing uncommitted changes")
        gd_doc = datasvc.gitsvc.GetGitDeploy(application, gd_name)
        branch = gd_doc['location']['default_branch']
        path = gd_doc['location']['path']
        res = rc.discard_git_changes(servers, path)
        logging.debug("_threadsafe_process_gitdeploy: discard git changes result: {}".format(str(res)))
        res = rc.checkout_branch(servers, path, branch)
        logging.debug("_threadsafe_process_gitdeploy: git checkout result: {}".format(str(res)))

        datasvc.jobsvc.NewJobData({"DeployServers": {gd_name: "deploying", "servers": servers}})
        logging.debug("_threadsafe_pull_gitdeploy: sls_map: {}".format(sls_map))

        datasvc.deploysvc.UpdateDeployment_Phase2(application, deployment_id, gd_name, servers, batch_number,
                                                  progress=50,
                                                  state="Issuing state commands (git pull, etc)")
        res = rc.run_slses_async(_threadsafe_pull_callback, sls_map, args={'datasvc': datasvc, 'application': application,
                                                                           'deployment_id': deployment_id,
                                                                           'batch_number': batch_number,
                                                                           'gitdeploy': gd_name})
        logging.debug("_threadsafe_pull_gitdeploy: results: {}".format(res))

        errors = dict()
        successes = dict()
        for r in res:
            for host in r:
                for cmd in r[host]['ret']:
                    if "gitdeploy" in cmd:
                        if "result" in r[host]['ret'][cmd]:
                            if not r[host]['ret'][cmd]["result"]:
                                errors[host] = r[host]['ret'][cmd]["changes"] if "changes" in r[host]['ret'][cmd] else r[host]['ret'][cmd]
                            else:
                                if host not in successes:
                                    successes[host] = dict()
                                module, state, command, subcommand = str(cmd).split('|')
                                if state not in successes[host]:
                                    successes[host][state] = dict()
                                successes[host][state][command] = {
                                    "stdout": r[host]['ret'][cmd]["changes"]["stdout"],
                                    "stderr": r[host]['ret'][cmd]["changes"]["stderr"],
                                    "retcode": r[host]['ret'][cmd]["changes"]["retcode"],
                                }
        if len(errors) > 0:
            for e in errors:
                datasvc.deploysvc.UpdateDeployment_Phase2(application, deployment_id, gd_name, [e], batch_number,
                                                  state="ERROR: {}".format(errors[e]))
            logging.debug("_threadsafe_pull_gitdeploy: SLS error servers: {}".format(errors.keys()))
            logging.debug("_threadsafe_pull_gitdeploy: SLS error responses: {}".format(errors))

        if len(successes) > 0:
            datasvc.deploysvc.UpdateDeployment_Phase2(application, deployment_id, gd_name, successes.keys(), batch_number,
                                                      progress=100, state="Complete")

        missing = list(set([host for r in res for host in r]).difference(set(servers)))
        if missing:
            datasvc.deploysvc.UpdateDeployment_Phase2(application, deployment_id, gd_name, missing, batch_number,
                                                      state="ERROR: no results (timed out waiting for salt?)")
            logging.debug("_threadsafe_pull_gitdeploy: error: empty results for: {}; possible salt timeout".format(missing))
            datasvc.jobsvc.NewJobData({"_threadsafe_pull_gitdeploy": "empty results for {}".format(missing)})
            datasvc.deploysvc.FailDeployment(application, deployment_id)

        deploy_results = {
            gd_name: {
                "raw_results": res,
                "errors": len(errors) > 0,
                "error_results": errors,
                "successes": len(successes) > 0,
                "success_results": successes
            }
        }

        queue.put_nowait(deploy_results)

        datasvc.jobsvc.NewJobData({
            "DeployServers": deploy_results
        })

        logging.debug("_threadsafe_pull_gitdeploy: finished ({})".format(gitdeploy_struct))

    except:
        exc_type, exc_obj, tb = sys.exc_info()
        exc_msg = "ERROR: Exception in _threadsafe_pull_gitdeploy: {}".format(traceback.format_exception(exc_type, exc_obj, tb))
        datasvc.deploysvc.UpdateDeployment_Phase2(application, deployment_id, gd_name, servers, batch_number,
                                                  state=exc_msg)
        datasvc.deploysvc.FailDeployment(application, deployment_id)


class DeployController:
    '''
    Class that runs deploys. Only knows about server/gitdeploy pairs, so is used for both manual-style deployments
    and group/environment deployments.
    '''

    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, datasvc, deployment_id):
        self.deployment_id = deployment_id
        self.datasvc = datasvc
        self.changed_gitdeploys = dict()

    def run(self, app_name, build_name, servers, gitdeploys, force=False, parallel=True, batch_number=0):
        '''
        1. Decompress build to gitdeploy dir and push
            a. Attempts to optimize by determining if build has already been decompressed to gitdeploy and skips if so
        2. Determine which gitdeploys have changes (if any)
            a. Build a mapping of gitdeploys_with_changes -> [ servers_to_deploy_it_to ]
            b. Perform the state calls only to the server/gitdeploy pairs that have changes

        @type app_name: str
        @type build_name: str
        @type servers: list(str)
        @type gitdeploys: list(str)
        '''
        assert app_name and build_name and servers and gitdeploys
        assert elita.util.type_check.is_string(app_name)
        assert elita.util.type_check.is_string(build_name)
        assert elita.util.type_check.is_seq(servers)
        assert elita.util.type_check.is_seq(gitdeploys)
        assert isinstance(batch_number, int) and batch_number >= 0

        build_doc = self.datasvc.buildsvc.GetBuildDoc(app_name, build_name)
        gitdeploy_docs = {gd: self.datasvc.gitsvc.GetGitDeploy(app_name, gd) for gd in gitdeploys}
        #reset changed gitdeploys
        self.changed_gitdeploys = dict()

        queue = billiard.Queue()
        procs = list()

        #we need to get a list of gitdeploys with unique gitrepos, so build a reverse mapping
        gitrepo_gitdeploy_mapping = {gitdeploy_docs[gd]['location']['gitrepo']['name']: gd for gd in gitdeploys}

        self.datasvc.deploysvc.StartDeployment_Phase(app_name, self.deployment_id, 1)
        for gr in gitrepo_gitdeploy_mapping:
            gd = gitrepo_gitdeploy_mapping[gr]
            gddoc = gitdeploy_docs[gd]
            if parallel:
                p = billiard.Process(target=_threadsafe_process_gitdeploy, name=gd,
                                     args=(gddoc, build_doc, servers, queue, self.datasvc.settings,
                                     self.datasvc.job_id, self.deployment_id))
                p.start()
                procs.append(p)

            else:
                _threadsafe_process_gitdeploy(gddoc, build_doc, servers, queue, self.datasvc.settings,
                                              self.datasvc.job_id, self.deployment_id)

            # if this is part of a rolling deployment and is anything other than the first batch,
            # no gitdeploys will actually be "changed". Force says to do a pull on the servers anyway.
            if force:
                logging.debug("Force flag set, adding gitdeploy servers to deploy list: {}".format(gd))
                queue.put_nowait({gd: determine_deployabe_servers(gddoc['servers'], servers)})

        if parallel:
            for p in procs:
                p.join(300)
                if p.is_alive():
                    p.terminate()
                    logging.debug("ERROR: _threadsafe_process_gitdeploy: timeout waiting for child process ({})!".
                                  format(p.name))
                    self.datasvc.jobsvc.NewJobData({'status': 'error',
                                                    'message': 'timeout waiting for child process (process_gitdeploy: {}'.format(p.name)})
                    self.datasvc.deploysvc.UpdateDeployment_Phase1(app_name, self.deployment_id, p.name,
                                                          step="ERROR: timed out waiting for child process")
                    self.datasvc.deploysvc.FailDeployment(app_name, self.deployment_id)
                    return False, None


        while not queue.empty():
            gd = queue.get_nowait()
            if gd:
                for g in gd:
                    self.changed_gitdeploys[g] = gd[g]

        logging.debug("DeployController: run: post-phase1: changed_gitdeploys: {}".format(self.changed_gitdeploys))

        queue = billiard.Queue()
        procs = list()
        self.datasvc.deploysvc.StartDeployment_Phase(app_name, self.deployment_id, 2)
        for gd in self.changed_gitdeploys:
            if parallel:
                p = billiard.Process(target=_threadsafe_pull_gitdeploy, name=gd,
                                            args=(app_name, {gd: self.changed_gitdeploys[gd]}, queue, self.datasvc.settings,
                                                  self.datasvc.job_id, self.deployment_id, batch_number))
                p.start()
                procs.append(p)
            else:
                _threadsafe_pull_gitdeploy(app_name, {gd: self.changed_gitdeploys[gd]}, queue, self.datasvc.settings,
                                           self.datasvc.job_id, self.deployment_id, batch_number)
        if parallel:
            for p in procs:
                p.join(300)
                if p.is_alive():
                    p.terminate()
                    logging.debug("ERROR: _threadsafe_pull_gitdeploy: timeout waiting for child process ({})!".
                                  format(p.name))
                    self.datasvc.jobsvc.NewJobData({'status': 'error',
                                                    'message': 'timeout waiting for child process (pull_gitdeploy: {}'.format(p.name)})
                    self.datasvc.deploysvc.UpdateDeployment_Phase2(app_name, self.deployment_id, p.name,
                                                                   self.changed_gitdeploys[p.name], batch_number,
                                                                   state="ERROR: timeout waiting for child process")
                    self.datasvc.deploysvc.FailDeployment(app_name, self.deployment_id)
                    return False, None

        results = list()
        while not queue.empty():
            results.append(queue.get_nowait())

        if not results:
            return False, results

        for r in results:
            for gd in r:
                if r[gd]['errors']:
                    return False, results

        #update deployed_build
        for gd in gitdeploys:
            gdm = gitservice.GitDeployManager(gitdeploy_docs[gd], self.datasvc)
            gdm.update_last_deployed(build_name)

        return True, results




import requests
from slugify import slugify
import tempfile
import os
import shutil
import sh
import git
import copy
import socket
import logging

import elita.util
import elita.util.type_check
import salt_control
import elita.builds
import sshconfig

__author__ = 'bkeroack'

class ValidRepoTypes:
    type_names = ['github', 'bitbucket']

class BitBucketData:
    api_base_url = 'https://api.bitbucket.org/2.0'

class GitHubData:
    api_base_url = 'https://github.com'

#mongo does not allow dots in key names, so we have to replace dots in embedded yaml dict keys with this
EMBEDDED_YAML_DOT_REPLACEMENT = '#'

#callables for async execution
def create_bitbucket_repo(datasvc, gitprovider, name, application, keypair):
    bbsvc = BitBucketRepoService(gitprovider, datasvc.settings)
    resp = bbsvc.create_repo(name)
    uri = bbsvc.get_ssh_uri(name)
    if uri:
        logging.debug("got uri: {}".format(uri))
        alias = bbsvc.key_setup(name, application, keypair)
        alias_uri = uri.replace("bitbucket.org", alias)
        logging.debug("alias uri: {}".format(alias_uri))
        bbsvc.setup_gitdeploy_dir(name, application, alias_uri, empty=True)
        datasvc.gitsvc.UpdateGitRepo(application, name, {'uri': uri})
    else:
        logging.debug("ERROR: uri for gitrepo not found!")
        resp['error'] = {'message': 'error getting uri!, local master not initialized!'}
    return resp

def create_github_repo(datasvc, gitprovider, name, application, keypair):
    return {'error': 'not implemented'}

def create_repo_callable_from_type(repo_type):
    if repo_type not in ValidRepoTypes.type_names:
        return None
    return create_bitbucket_repo if repo_type == 'bitbucket' else create_github_repo

def delete_repo_callable_from_type(repo_type):
    if repo_type not in ValidRepoTypes.type_names:
        return None
    return delete_bitbucket_repo if repo_type == 'bitbucket' else delete_github_repo

def delete_bitbucket_repo(datasvc, gitprovider, name):
    bbrs = BitBucketRepoService(gitprovider, datasvc.settings)
    return bbrs.delete_repo(name)

def delete_github_repo(datasvc, gitprovider, name):
    return {'error': 'not implemented'}

def create_gitdeploy(datasvc, gitdeploy):
    gdm = GitDeployManager(gitdeploy, datasvc)
    gdm.add_sls()
    return "done"

def remove_gitdeploy(datasvc, gitdeploy):
    gdm = GitDeployManager(gitdeploy, datasvc)
    gdm.rm_sls()
    datasvc.gitsvc.DeleteGitDeploy(gitdeploy['application'], gitdeploy['name'])
    return "done"

def initialize_gitdeploy(datasvc, gitdeploy, server_list):
    sc = salt_control.SaltController(datasvc)
    for s in server_list:
        if not sc.verify_connectivity(s):
            return {'error': "server '{}' not accessible via salt".format(s)}
    gdm = GitDeployManager(gitdeploy, datasvc)
    res = gdm.initialize(server_list)
    if 'servers' not in gitdeploy:
        gitdeploy['servers'] = list()
    sset = set(tuple(gitdeploy['servers']))
    for s in server_list:
        sset.add(s)
    datasvc.gitsvc.UpdateGitDeploy(gitdeploy['application'], gitdeploy['name'], {'servers': list(sset)})
    return res

def deinitialize_gitdeploy(datasvc, gitdeploy, server_list):
    sc = salt_control.SaltController(datasvc)
    for s in server_list:
        if not sc.verify_connectivity(s):
            return {'error': "server '{}' not accessible via salt".format(s)}
    gdm = GitDeployManager(gitdeploy, datasvc)
    res = gdm.deinitialize(server_list)
    existing_servers = gitdeploy['servers'] if 'servers' in gitdeploy else list()
    for s in server_list:
        if s in existing_servers:
            existing_servers.remove(s)
    datasvc.gitsvc.UpdateGitDeploy(gitdeploy['application'], gitdeploy['name'], {'servers': existing_servers})
    return res

def remove_and_deinitialize_gitdeploy(datasvc, gitdeploy):
    assert 'servers' in gitdeploy
    server_list = gitdeploy['servers']
    res1 = deinitialize_gitdeploy(datasvc, gitdeploy, server_list)
    res2 = remove_gitdeploy(datasvc, gitdeploy)
    return {
        'deinitialize': res1,
        'remove': res2
    }

def setup_local_gitrepo_dir(datasvc, gitrepo):
    logging.debug("setup_local_gitrepo: {}".format(gitrepo['name']))
    repo_service = BitBucketRepoService if gitrepo['gitprovider']['type'] == 'bitbucket' else GitHubRepoService
    rs = repo_service(gitrepo['gitprovider'], datasvc.settings)
    kp = gitrepo['keypair']
    rs.key_setup(gitrepo['name'], gitrepo['application'], kp)
    rs.setup_gitdeploy_dir(gitrepo['name'], gitrepo['application'], gitrepo['uri'], empty=False)

class GitRepoService:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, gitprovider, settings):
        self.gp_type = gitprovider['type']
        self.settings = settings
        self.auth = gitprovider['auth']
        self.base_url = BitBucketData.api_base_url if self.gp_type == 'bitbucket' else GitHubData.api_base_url

    def create_repo(self, name):
        logging.debug("create_repo not implemented")

    def get_alias(self, gitrepo_name, app):
        sshc = sshconfig.SSHController()
        return sshc.get_alias(gitrepo_name, app)

    def get_alias_uri(self, alias, original_uri):
        orig_domain = 'bitbucket.org' if self.gp_type == 'bitbucket' else 'github.com'
        return original_uri.replace(orig_domain, alias)

    def key_setup(self, gitrepo_name, application, keypair):
        '''
        copy keypair to user ssh dir
        add alias in ~/.ssh/config
        '''
        sshc = sshconfig.SSHController()
        alias_name = sshc.write_local_keys(application, self.gp_type, gitrepo_name, keypair['private_key'],
                                           keypair['public_key'])
        return alias_name

    def git_user_config(self, cwd):
        git = sh.git.bake(_cwd=cwd)
        hostname = socket.getfqdn()
        res = git.config("user.email", "elita@{}".format(hostname))
        logging.debug("setup_gitdeploy_dir: git config user.email: {}".format(res))
        res = git.config("user.name", "elita")
        logging.debug("setup_gitdeploy_dir: git config user.name: {}".format(res))
        res = git.config("--global", "push.default", "simple")
        logging.debug("setup_gitdeploy_dir: git config --global push.default simple: {}".format(res))

    def setup_gitdeploy_dir(self, gitrepo_name, application, uri, empty=False):
        '''create master gitdeploy directory and initialize with git'''
        logging.debug("setup_gitdeploy_dir: name: {}; app: {}; uri: {}".format(gitrepo_name, application, uri))
        root = self.settings['elita.gitdeploy.dir']
        if not os.path.isdir(root):
            os.mkdir(root)
        parent_path = os.path.join(root, application)
        if not os.path.isdir(parent_path):
            os.mkdir(parent_path)
        path = os.path.join(parent_path, gitrepo_name)
        alias = self.get_alias(gitrepo_name, application)
        alias_uri = self.get_alias_uri(alias , uri)
        if os.path.isdir(path) and empty:
            shutil.rmtree(path)
        if not os.path.isdir(path):
            os.mkdir(path)
            if empty:
                git = sh.git.bake(_cwd=path)
                res = git.init()
                logging.debug("setup_gitdeploy_dir: git init: {}".format(res))
                res = git.remote.add.origin("ssh://{}".format(alias_uri))
                logging.debug("setup_gitdeploy_dir: git remote add origin: {}".format(res))
                logging.debug("setup_gitdeploy_dir: creating initial repo with dummy file")
                touch = sh.touch.bake(_cwd=path)
                touch(".empty")
                res = git.add('-A')
                logging.debug("setup_gitdeploy_dir: git add: {}".format(res))
                self.git_user_config(path)
                res = git.commit(m="initial state")
                logging.debug("setup_gitdeploy_dir: git commit: {}".format(res))
                res = git.push("--set-upstream", "origin", "master")
                logging.debug("setup_gitdeploy_dir: git push: {}".format(res))
            else:
                git = sh.git.bake(_cwd=parent_path)
                logging.debug("setup_gitdeploy_dir: cloning repo")
                res = git.clone("ssh://{}".format(alias_uri), gitrepo_name)
                logging.debug("setup_gitdeploy_dir: res: {}".format(res))
                res = git.branch("--set-upstream-to=origin/master", "master")
                logging.debug("setup_gitdeploy_dir: git set upstream: {}".format(res))
                self.git_user_config(path)
        else:
            logging.debug("setup_gitdeploy_dir: local dir exists! not creating")

class GitHubRepoService(GitRepoService):
    __metaclass__ = elita.util.LoggingMetaClass
    pass

class BitBucketRepoService(GitRepoService):
    __metaclass__ = elita.util.LoggingMetaClass

    def create_repo(self, name):
        slug = slugify(name)
        username = self.auth['username']
        password = self.auth['password']
        json_body = {
            'scm': 'git',
            'name': name,
            'is_private': 'true',
            'description': "Created by elita",
            'forking_policy': 'no_forks'
        }
        r = requests.post("{}/repositories/{}/{}".format(self.base_url, username, slug), data=json_body,
                          auth=(username, password))
        try:
            resp = r.json()
        except:
            resp = r.text
        return resp

    def delete_repo(self, name):
        logging.debug("Deleting repo: {}".format(name))
        slug=slugify(name)
        username = self.auth['username']
        password = self.auth['password']
        r = requests.delete("{}/repositories/{}/{}".format(self.base_url, username, slug), auth=(username, password))
        try:
            resp = r.json()
        except:
            resp = r.text
        return resp

    def get_ssh_uri(self, name):
        logging.debug("getting clone URI for {}".format(name))
        slug = slugify(name)
        username = self.auth['username']
        password = self.auth['password']
        r = requests.get("{}/repositories/{}/{}".format(self.base_url, username, slug), auth=(username, password))
        try:
            resp = r.json()
        except:
            logging.debug("exception parsing json resp!")
            return None
        for l in resp['links']['clone']:
            if l['name'] == "ssh":
                return l['href'][6:]  # chop off leading 'ssh://'
        logging.debug("nothing found in response!")
        return None

class GitDeployManager:
    __metaclass__ = elita.util.LoggingMetaClass
    
    def __init__(self, gitdeploy, datasvc):
        '''
        @type gitdeploy: dict
        @type datasvc: elita.models.DataService
        '''
        assert gitdeploy and datasvc
        assert elita.util.type_check.is_dictlike(gitdeploy)
        self.datasvc = datasvc
        self.gitdeploy = gitdeploy
        self.settings = datasvc.settings
        self.sc = salt_control.SaltController(datasvc)
        self.rc = salt_control.RemoteCommands(self.sc)

        if not os.path.isdir(self.get_path()):
            gitrepo = gitdeploy['location']['gitrepo']
            setup_local_gitrepo_dir(datasvc, gitrepo)

        #if there's no uri for associated gitrepo, try to fetch it
        if not gitdeploy['location']['gitrepo']['uri']:
            logging.debug("WARNING: found gitrepo with empty URI; fixing")
            repo_service = BitBucketRepoService if \
                gitdeploy['location']['gitrepo']['gitprovider']['type'] == "bitbucket" else GitHubRepoService
            rs = repo_service(gitdeploy['location']['gitrepo']['gitprovider'], self.settings)
            uri = rs.get_ssh_uri(gitdeploy['location']['gitrepo']['name'])
            assert uri is not None
            #need to get the original doc because of the dereferences
            git_repo = datasvc.gitsvc.GetGitRepo(gitdeploy['application'], gitdeploy['location']['gitrepo']['name'])
            datasvc.gitsvc.UpdateGitRepo(gitdeploy['application'], git_repo['name'], {'uri': uri})
            self.gitdeploy = datasvc.gitsvc.GetGitDeploy(gitdeploy['application'], gitdeploy['name'])

        if "last_build" in gitdeploy['location']['gitrepo']:
            self.last_build = gitdeploy['location']['gitrepo']['last_build']
        else:
            logging.debug("WARNING: found gitrepo without last_build")
            self.last_build = None

        self.deployed_build = gitdeploy['deployed_build']
        self.stale = False

    def update_stale(self):
        self.gitdeploy = self.datasvc.gitsvc.GetGitDeploy(self.gitdeploy['application'], self.gitdeploy['name'])
        self.last_build = self.gitdeploy['location']['gitrepo']['last_build']
        self.deployed_build = self.gitdeploy['deployed_build']
        self.stale = self.last_build != self.deployed_build

    def initialize(self, server_list):
        return {
            'prehook': self.run_init_prehook(server_list),
            'delete_remote_dir': self.delete_remote_dir(server_list),
            'create_remote_dir': self.create_remote_dir(server_list),
            'push_keys': self.push_keypair(server_list),
            'clone_repo': self.clone_repo(server_list),
            'create_gitignore': self.create_ignore(server_list),
            'posthook': self.run_init_posthook(server_list)
        }

    def deinitialize(self, server_list):
        return {
            'prehook': self.run_deinit_prehook(server_list),
            'delete_remote_dir': self.delete_remote_dir(server_list),
            'delete_remote_keys': self.delete_remote_keypair(server_list),
            'posthook': self.run_deinit_posthook(server_list)
        }

    def run_hook(self, hook, **kwargs):
        self.datasvc.jobsvc.NewJobData({'status': 'running hook {}'.format(hook)})
        args = {
            'hook_parameters':
                {
                    'gitdeploy': self.gitdeploy
                }
        }
        for k in kwargs:
            args['hook_parameters'][k] = kwargs[k]
        return self.datasvc.actionsvc.hooks.run_hook(self.gitdeploy['application'], hook, args)

    def run_init_prehook(self, server_list):
        return self.run_hook("GITDEPLOY_INIT_PRE", server_list=server_list)

    def run_init_posthook(self, server_list):
        return self.run_hook("GITDEPLOY_INIT_POST", server_list=server_list)

    def run_deinit_prehook(self, server_list):
        return self.run_hook("GITDEPLOY_DEINIT_PRE", server_list=server_list)

    def run_deinit_posthook(self, server_list):
        return self.run_hook("GITDEPLOY_DEINIT_POST", server_list=server_list)

    def run_commit_diffhook(self, files):
        return self.run_hook("GITDEPLOY_COMMIT_DIFF", files=files)

    def add_sls(self):
        self.sc.new_gitdeploy_yaml(self.gitdeploy)

    def rm_sls(self):
        self.sc.rm_gitdeploy_yaml(self.gitdeploy)

    def delete_remote_dir(self, server_list):
        logging.debug("delete_remote_dir")
        path = self.gitdeploy['location']['path']
        res = self.rc.delete_directory(server_list, path)
        logging.debug("delete_remote_dir on servers: {}: resp: {}".format(server_list, res))
        return res

    def create_remote_dir(self, server_list):
        logging.debug("create_remote_dir")
        path = self.gitdeploy['location']['path']
        res = self.rc.create_directory(server_list, path)
        logging.debug('create_remote_dir on servers: {}: resp: {}'.format(server_list, res))
        return res

    def delete_remote_keypair(self, server_list):
        return {'status': 'noop'}

    def push_keypair(self, server_list):
        '''
        Push the keypair to the target servers and set up SSH alias
        '''
        sshc = sshconfig.SSHController()
        return sshc.push_remote_keys(self.rc, server_list, self.gitdeploy['application'],
                                     self.gitdeploy['location']['gitrepo']['name'],
                                     self.gitdeploy['location']['gitrepo']['gitprovider']['type'],
                                     self.gitdeploy['location']['gitrepo']['keypair']['private_key'],
                                     self.gitdeploy['location']['gitrepo']['keypair']['public_key'])

    def clone_repo(self, server_list):
        logging.debug("clone_repo: cloning")
        sshc = sshconfig.SSHController()
        uri = sshc.get_full_alias_uri(self.gitdeploy['location']['gitrepo'])
        uri = "ssh://{}".format(uri)
        dest = self.gitdeploy['location']['path']
        res = self.rc.clone_repo(server_list, uri, dest)
        logging.debug("clone_repo: resp: {}".format(res))
        return res

    def create_ignore(self, server_list):
        repo_path = self.gitdeploy['location']['path']
        fd, temp_name = tempfile.mkstemp()
        logging.debug("create_ignore: writing file")
        if 'gitignore' in self.gitdeploy['options']:
            gi = self.gitdeploy['options']['gitignore']
            assert isinstance(gi, list)
            with open(temp_name, 'w') as f:
                for l in gi:
                    f.write("{}\n".format(l))
        logging.debug("create_ignore: pushing gitignore")
        remote_filename = os.path.join(repo_path, ".gitignore")
        res = self.rc.push_files(server_list, {temp_name: remote_filename})
        logging.debug("create_ignore: push resp: {}".format(res))
        logging.debug("create_ignore: autocrlf")
        res = self.rc.set_git_autocrlf(server_list, repo_path)
        logging.debug("create_ignore: autocrlf resp: {}".format(res))
        logging.debug("create_ignore: disabling push")
        res = self.rc.set_git_push_url(server_list, repo_path, "do.not.push")
        logging.debug("create_ignore: disable push resp: {}".format(res))
        logging.debug("create_ignore: adding file")
        res = self.rc.add_all_files_git(server_list, repo_path)
        logging.debug("create_ignore: add resp: {}".format(res))
        logging.debug("create_ignore: setting user config")
        res = self.rc.set_user_git(server_list, repo_path, "elita", "elita@elitaserver")
        logging.debug("create_ignore: config resp: {}".format(res))
        logging.debug("create_ignore: committing")
        res = self.rc.commit_git(server_list, repo_path, "gitignore")
        logging.debug("create_ignore: commit resp: {}".format(res))
        return res

    def get_path(self):
        appname = self.gitdeploy['application']
        gitrepo_name = self.gitdeploy['location']['gitrepo']['name']
        root = self.settings['elita.gitdeploy.dir']
        return "{}/{}/{}".format(root, appname, gitrepo_name)

    def git_obj(self):
        path = self.get_path()
        return sh.git.bake(_cwd=path)

    def inspect_latest_diff(self):
        repo = git.Repo(self.get_path())
        m = repo.heads.master
        logging.debug("commit hash: {}".format(m.commit.hexsha))
        s = m.commit.stats
        #replace all '.' with underscore to make Mongo happy
        #make copy for the hook
        return s.files    # not appropriate for direct inclusion into mongo due to filenames as keys

    def checkout_default_branch(self):
        branch = self.gitdeploy['location']['default_branch']
        logging.debug("checkout_default_branch: git checkout {}".format(branch))
        git = self.git_obj()
        return git.checkout(branch)

    def decompress_to_repo(self, package_doc):
        path = self.get_path()
        logging.debug("decompress_to_repo: deleting contents")
        for f in os.listdir(path):
            fpath = os.path.join(path, f)
            if os.path.isdir(fpath) and ".git" not in f:
                shutil.rmtree(fpath)
            elif os.path.isfile(fpath):
                os.unlink(fpath)
        logging.debug("decompress_to_repo: decompressing {} to {}".format(package_doc['filename'], path))
        bf = elita.builds.BuildFile(package_doc)
        bf.decompress(path)

    def check_repo_status(self):
        git = self.git_obj()
        return git.status()

    def add_files_to_repo(self):
        logging.debug("add_files_to_repo: git add")
        git = self.git_obj()
        return git.add('-A')

    def commit_to_repo(self, build_name):
        logging.debug("commit_to_repo: git commit")
        git = self.git_obj()
        return git.commit(m=build_name)

    def push_repo(self):
        logging.debug("push_repo: git push")
        git = self.git_obj()
        return git.push("-u", "origin", self.gitdeploy['location']['default_branch'])

    def update_repo_last_build(self, build_name):
        gitrepo_name = self.gitdeploy['location']['gitrepo']['name']
        logging.debug("update_repo_last_build: updating last_build on {} to {}".format(gitrepo_name, build_name))
        gitrepo = self.datasvc.gitsvc.GetGitRepo(self.gitdeploy['application'], gitrepo_name)
        gitrepo['last_build'] = build_name
        self.datasvc.gitsvc.UpdateGitRepo(self.gitdeploy['application'], gitrepo_name, gitrepo)
        self.update_stale()

    def update_last_deployed(self, build_name):
        self.datasvc.gitsvc.UpdateGitDeploy(self.gitdeploy['application'], self.gitdeploy['name'],
                                            {'deployed_build': build_name})
        self.update_stale()





__author__ = 'bkeroack'

import os
import pkg_resources

from clint.arguments import Args
from clint.textui import puts, colored

import subcommands

config = {
    'host': 'localhost' if 'ELITA_HOST' not in os.environ else os.environ['ELITA_HOST'],
    'port': 2718 if 'ELITA_PORT' not in os.environ else os.environ['ELITA_PORT'],
    'secure': 'ELITA_SECURE' in os.environ,
    'ignore_cert': 'ELITA_IGNORE_CERT' in os.environ
}

sub_commands = {
    'about': subcommands.About
}

help_text = """
{title} {version}
=================

USAGE: elita [SUBCOMMAND] [OPTIONS]

SUBCOMMANDS:

{subcommands}
""".format(title=colored.green('elita'), version=colored.magenta(pkg_resources.require("elita")[0].version),
           subcommands="\n".join([k for k in sub_commands]))

def Command_Line_Client():

    args = Args().grouped

    for item in args:
        if item is '_':
            sc = args[item].all
            if not sc or len(sc) != 1 or sc[0] not in sub_commands:
                puts(help_text)
            else:
                sub_commands[sc[0]](config)


import logging
import os
import stat
import elita.util
import lockfile
import tempfile

import elita.util.type_check
from salt_control import OSTypes, RemoteCommands

ELITA_USER = "elita"
#remote key locations
WINDOWS_KEY_LOCATION = 'C:\Program Files (x86)\Git\.ssh'
UNIX_KEY_LOCATION = '~/.ssh'
UNIX_MINION_USER = 'root'  # hack

class UnknownGitproviderType(Exception):
    pass

class SSHController:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self):
        pass

    def get_alias_block(self, alias_name, real_hostname, keyfile, username='git'):
        '''
        Return the ssh config block for the requested alias. We use tabs to make it easier to parse if we ever need to
        in the future.
        '''
        return """
Host {alias}
\tHostname {hostname}
\tUser {username}
\tPreferredAuthentications publickey
\tStrictHostKeyChecking no
\tIdentityFile "{key}"

""".format(alias=alias_name, hostname=real_hostname, username=username, key=keyfile)

    def get_local_ssh_dir(self):
        '''
        Get path of local .ssh config and create it if necessary
        '''
        home_dir = os.path.expanduser('~{}'.format(ELITA_USER))
        home_sshdir = os.path.join(home_dir, ".ssh")
        if not os.path.isdir(home_sshdir):
            os.mkdir(home_sshdir)
        return home_sshdir

    def get_key_name(self, application, gitrepo_name):
        return "{}-{}".format(application, gitrepo_name)

    def get_alias(self, gitrepo_name, app):
        return "{}-{}".format(app, gitrepo_name)

    def get_full_alias_uri(self, gitrepo):
        '''
        Get the full aliased URI for a given gitrepo
        '''
        assert gitrepo
        assert elita.util.type_check.is_dictlike(gitrepo)
        assert 'gitprovider' in gitrepo
        assert elita.util.type_check.is_dictlike(gitrepo['gitprovider'])
        assert 'type' in gitrepo['gitprovider']     # make sure gitprovider is dereferenced

        alias_host = self.get_alias(gitrepo['name'], gitrepo['application'])
        if gitrepo['gitprovider']['type'] == 'bitbucket':
            return gitrepo['uri'].replace('bitbucket.org', alias_host)
        elif gitrepo['gitprovider']['type'] == 'github':
            return gitrepo['uri'].replace('github.com', alias_host)
        else:
            raise UnknownGitproviderType

    def add_local_alias(self, alias_block):
        '''
        Add ssh alias (for gitdeploy most likely) to the local ssh config
        '''
        ssh_config = os.path.join(self.get_local_ssh_dir(), "config")
        lock = lockfile.FileLock(ssh_config)
        lock.acquire(timeout=60)
        with open(ssh_config, 'a') as f:
            f.write(alias_block)
        os.chmod(ssh_config, stat.S_IWUSR | stat.S_IRUSR)
        lock.release()

    def add_remote_alias(self, remote_controller, server_list, sshconf, alias_block):
        '''
        Add ssh alias to remote ssh config via file.append

        This is pretty hacky since the alias could already exist and this will just append to the config endlessly,
        but fixing that would involve parsing the whole config and that gets complicated.
        '''
        remote_controller.touch(server_list, sshconf)
        return remote_controller.append_to_file(server_list, sshconf, alias_block)


    def write_local_keys(self, application, gitrepo_type, gitrepo_name, private_key_data, public_key_data):
        '''
        Write key data to named key files for gitrepo
        '''
        home_sshdir = self.get_local_ssh_dir()
        keyname = self.get_key_name(application, gitrepo_name)
        priv_key_name = os.path.join(home_sshdir, keyname)
        pub_key_name = "{}.pub".format(priv_key_name)

        logging.debug("write_local_keys: writing keypairs")
        with open(pub_key_name, 'w') as f:
            f.write(public_key_data.decode('string_escape'))

        with open(priv_key_name, 'w') as f:
            f.write(private_key_data.decode('string_escape'))

        logging.debug("write_local_keys: chmod private key to owner read/write only")
        os.chmod(priv_key_name, stat.S_IWUSR | stat.S_IRUSR)

        logging.debug("write_local_keys: adding alias to ssh config")
        alias_name = self.get_alias(gitrepo_name, application)
        alias_block = self.get_alias_block(alias_name, "bitbucket.org" if gitrepo_type == 'bitbucket' else 'github.com',
                                           priv_key_name)
        self.add_local_alias(alias_block)
        return alias_name

    def create_key_dir(self, remote_controller, server_list):
        '''
        Create platform-specific directory to hold ssh config and keys
        @type remote_controller: RemoteCommands
        '''
        assert remote_controller and server_list
        assert isinstance(remote_controller, RemoteCommands)
        assert elita.util.type_check.is_seq(server_list)

        servers_by_os = remote_controller.group_remote_servers_by_os(server_list)
        results = {'windows': dict(), 'unix': dict()}
        if len(servers_by_os['windows']) > 0:
            results['windows'] = remote_controller.create_directory(servers_by_os['windows'], WINDOWS_KEY_LOCATION)
        if len(servers_by_os['unix']) > 0:
            results['unix'] = remote_controller.create_directory(servers_by_os['unix'], UNIX_KEY_LOCATION)
        #merge results
        return dict({s: results['windows'][s] for s in results['windows']}, **{s: results['unix'][s] for s in results['unix']})

    def push_remote_keys(self, remote_controller, server_list, application, gitrepo_name, gitrepo_type, privkey_data, pubkey_data):
        '''
        Write keydata to temp files, push to remote servers. Add alias to remote ssh config
        '''

        f_pub, tf_pub = tempfile.mkstemp(text=True)
        with open(tf_pub, 'w') as f:
            f.write(pubkey_data.decode('string_escape'))

        f_priv, tf_priv = tempfile.mkstemp(text=True)
        with open(tf_priv, 'w') as f:
            f.write(privkey_data.decode('string_escape'))

        keyname = self.get_key_name(application, gitrepo_name)
        real_hostname = "bitbucket.org" if gitrepo_type == "bitbucket" else "github.com"
        servers_by_os = remote_controller.group_remote_servers_by_os(server_list)

        res_win = dict()
        if len(servers_by_os['windows']) > 0:
            local_to_remote_mapping = {
                tf_pub: "{}\{}.pub".format(WINDOWS_KEY_LOCATION, keyname),
                tf_priv: "{}\{}".format(WINDOWS_KEY_LOCATION, keyname)
            }
            alias_block = self.get_alias_block(self.get_alias(gitrepo_name, application), real_hostname,
                                               local_to_remote_mapping[tf_priv])
            sshconf = "{}\\config".format(WINDOWS_KEY_LOCATION)
            res_win['push_files'] = remote_controller.push_files(servers_by_os['windows'], local_to_remote_mapping)
            res_win['add_alias'] = self.add_remote_alias(remote_controller, servers_by_os['windows'], sshconf, alias_block)

        res_unix = dict()
        if len(servers_by_os['unix']) > 0:
            local_to_remote_mapping = {
                tf_pub: os.path.join(UNIX_KEY_LOCATION, "{}.pub".format(keyname)),
                tf_priv: os.path.join(UNIX_KEY_LOCATION, keyname)
            }
            alias_block = self.get_alias_block(self.get_alias(gitrepo_name, application), real_hostname,
                                               local_to_remote_mapping[tf_priv])
            sshconf = os.path.join(UNIX_KEY_LOCATION, "config")
            res_unix['push_files'] = remote_controller.push_files(servers_by_os['unix'], local_to_remote_mapping,
                                                                  mode="600")
            res_unix['add_alias'] = self.add_remote_alias(remote_controller, servers_by_os['unix'], sshconf, alias_block)

        os.unlink(tf_pub)
        os.unlink(tf_priv)

        return dict(res_win, **res_unix)